from __future__ import annotations

import asyncio
import datetime
import glob
import inspect
import itertools
import json
import os
import typing
from enum import Enum
from hashlib import md5
from pathlib import Path
from types import GenericAlias
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Union,
    get_args,
    get_origin,
)

import aiofiles
import dateparser
from deepdiff.model import PrettyOrderedSet
from git import Repo
from jinja2 import BaseLoader, Environment
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Extra, Field, root_validator, schema, validate_model, validator
from pydantic.fields import ModelField

from iambic.core.iambic_enum import Command, ExecutionStatus, IambicManaged
from iambic.core.logger import log
from iambic.core.utils import (
    apply_to_provider,
    create_commented_map,
    get_writable_directory,
    sanitize_string,
    simplify_dt,
    snake_to_camelcap,
    sort_dict,
    transform_comments,
    yaml,
)

if TYPE_CHECKING:
    from iambic.config.dynamic_config import Config
    from iambic.plugins.v0_1_0.aws.models import AWSAccount

    MappingIntStrAny = typing.Mapping[int | str, Any]
    AbstractSetIntStr = typing.AbstractSet[int | str]


def field_schema(field: ModelField, **kwargs: Any) -> Any:
    """
    Used to hide a field from the OpenAPI schema. If
    the field has a field_info.extra key "hidden_from_schema" set to True,
    then this code will skip the field and prevent it from being added to
    the OpenAPI schema.
    """
    if field.field_info.extra.get("hidden_from_schema", False):
        raise schema.SkipField(f"{field.name} field is being hidden")
    else:
        return original_field_schema(field, **kwargs)


original_field_schema = schema.field_schema
schema.field_schema = field_schema


class IambicPydanticBaseModel(PydanticBaseModel):
    metadata_iambic_fields = Field(
        set(), description="metadata for iambic", exclude=True, hidden_from_schema=True
    )
    metadata_commented_dict: dict = Field(
        {}, description="yaml inline comments", hidden_from_schema=True
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        ancestors = inspect.getmro(type(self))
        for ancestor in ancestors:
            if getattr(ancestor, "iambic_specific_knowledge", None):
                self.metadata_iambic_fields = self.metadata_iambic_fields.union(
                    ancestor.iambic_specific_knowledge()
                )

    class Config:
        json_encoders = {Set: list}

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"metadata_commented_dict"}

    # https://github.com/pydantic/pydantic/issues/1864#issuecomment-1118485697
    # Use this to validate post creation
    def validate_model_afterward(self):
        values, fields_set, validation_error = validate_model(
            self.__class__, self.__dict__
        )
        if validation_error:
            raise validation_error
        try:
            object.__setattr__(self, "__dict__", values)
        except TypeError as e:
            raise TypeError(
                "Model values must be a dict; you may not have returned "
                + "a dictionary from a root validator"
            ) from e
        object.__setattr__(self, "__fields_set__", fields_set)


class BaseModel(IambicPydanticBaseModel):
    """
    A base model class that provides additional helper methods and
    configurations for other models used in IAMbic.
    """

    @classmethod
    def update_forward_refs(cls, **kwargs):
        kwargs["Union"] = Union
        super().update_forward_refs(**kwargs)

    class Config:
        alias_generator = snake_to_camelcap
        allow_population_by_field_name = True

    @classmethod
    def required_fields(cls) -> list[str]:
        return [
            field_name
            for field_name, field in cls.__dict__.get("__fields__", {}).items()
            if field != Optional
        ]

    @staticmethod
    def get_field_type(field: Any) -> Any:
        """
        Resolves the base field type for a model
        """
        field_type = field.type_ if isinstance(field, ModelField) else field
        if field_type == Optional:
            field_type = field_type[0]

        if (
            type(field_type) in [dict, GenericAlias, list, List, Set, set]
            or get_origin(field_type) == Union
        ) and (field_types := get_args(field_type)):
            return BaseModel.get_field_type(field_types[0])

        return field_type

    def get_attribute_val_for_account(
        self,
        aws_account: AWSAccount,
        attr: str,
        as_boto_dict: bool = True,
    ):
        """
        Retrieve the value of an attribute for a specific AWS account.

        :param aws_account: The AWSAccount object for which the attribute value should be retrieved.
        :param attr: The attribute name (supports nested attributes via dot notation, e.g., properties.tags).
        :param as_boto_dict: If True, the value will be transformed to a boto dictionary if applicable.
        :return: The attribute value for the specified AWS account.
        """
        # Support for nested attributes via dot notation. Example: properties.tags
        attr_val = self
        for attr_key in attr.split("."):
            attr_val = getattr(attr_val, attr_key)

        if as_boto_dict and hasattr(attr_val, "_apply_resource_dict"):
            return attr_val._apply_resource_dict(aws_account)
        elif not isinstance(attr_val, list):
            return attr_val

        matching_definitions = [
            val for val in attr_val if apply_to_provider(val, aws_account)
        ]
        if len(matching_definitions) == 0:
            # Fallback to the default definition
            field = self
            split_key = attr.split(".")
            if len(split_key) > 1:
                for key in split_key[:-1]:
                    field = getattr(field, key)
            return field.__fields__[split_key[-1]].default
        elif as_boto_dict:
            return [
                match._apply_resource_dict(aws_account)
                if hasattr(match, "_apply_resource_dict")
                else match
                for match in matching_definitions
            ]
        else:
            return matching_definitions

    def _apply_resource_dict(self, aws_account: AWSAccount = None) -> dict:
        exclude_keys = {
            "deleted",
            "expires_at",
            "included_accounts",
            "excluded_accounts",
            "included_orgs",
            "excluded_orgs",
            "owner",
            "template_type",
            "file_path",
            "metadata_iambic_fields",
            "metadata_commented_dict",
        }
        exclude_keys.update(self.exclude_keys)
        has_properties = hasattr(self, "properties")
        properties = getattr(self, "properties", self)
        if aws_account:
            resource_dict = {
                k: self.get_attribute_val_for_account(
                    aws_account,
                    f"properties.{k}" if has_properties else k,
                )
                for k in properties.__dict__.keys()
                if k not in exclude_keys
            }
            resource_dict = {
                k: v for k, v in resource_dict.items() if v is not None
            }  # using bool(v) will make it impossible to handle empty string tag value
        else:
            resource_dict = properties.dict(
                exclude=exclude_keys, exclude_none=True, exclude_unset=False
            )

        return {self.case_convention(k): v for k, v in resource_dict.items()}

    def apply_resource_dict(self, aws_account: AWSAccount) -> dict:
        response = self._apply_resource_dict(aws_account)
        variables = {var.key: var.value for var in aws_account.variables}
        variables["account_id"] = aws_account.account_id
        variables["account_name"] = aws_account.account_name
        if hasattr(self, "owner") and (owner := getattr(self, "owner", None)):
            variables["owner"] = owner

        rtemplate = Environment(loader=BaseLoader()).from_string(json.dumps(response))
        valid_characters_re = r"[\w_+=,.@-]"
        variables = {
            k: sanitize_string(v, valid_characters_re) for k, v in variables.items()
        }
        data = rtemplate.render(var=variables)
        return json.loads(data)

    async def remove_expired_resources(self):
        # Look at current model and recurse through submodules to see if it is a subclass of ExpiryModel
        # If it is, then call the remove_expired_resources method

        if issubclass(type(self), ExpiryModel):
            if hasattr(self, "expires_at") and self.expires_at:
                if self.expires_at < datetime.datetime.now(datetime.timezone.utc):
                    self.deleted = True
                    log.info("Expired resource found, marking for deletion")
                    return self
        for field_name in self.__fields__.keys():
            field_val = getattr(self, field_name)
            if isinstance(field_val, list):
                await asyncio.gather(
                    *[
                        elem.remove_expired_resources()
                        for elem in field_val
                        if isinstance(elem, BaseModel)
                    ]
                )
                for elem in field_val:
                    if getattr(elem, "deleted", None) is True:
                        field_val.remove(elem)

            elif not (
                isinstance(field_val, BaseModel) or isinstance(field_val, ExpiryModel)
            ):
                continue

            else:
                if isinstance(field_val, BaseModel):
                    await field_val.remove_expired_resources()
                    if getattr(field_val, "deleted", None) is True:
                        setattr(self, field_name, None)

    @property
    def exclude_keys(self) -> set:
        return set()

    @property
    def case_convention(self):
        return snake_to_camelcap

    @property
    def resource_type(self) -> str:
        raise NotImplementedError

    @property
    def resource_id(self) -> str:
        raise NotImplementedError


class ProposedChangeType(Enum):
    CREATE = "Create"
    UPDATE = "Update"
    DELETE = "Delete"
    ATTACH = "Attach"
    DETACH = "Detach"


class ProposedChange(PydanticBaseModel):
    change_type: ProposedChangeType
    account: Optional[
        str
    ]  # Used for Org related changes like permission set account assignments
    attribute: Optional[str]
    resource_id: Optional[str]
    resource_type: Optional[str]
    current_value: Optional[Union[list, dict, str, int, None]]
    new_value: Optional[Union[list, dict, str, int, None]]
    change_summary: Optional[dict]
    exceptions_seen: list[str] = Field(
        default=[]
    )  # FIXME, can we do better than string?


class AccountChangeDetails(PydanticBaseModel):
    org_id: Optional[str]
    account: Union[str, int]
    resource_id: Union[str, int]
    current_value: Optional[dict]
    new_value: Optional[dict]
    proposed_changes: list[ProposedChange] = Field(default=[])
    exceptions_seen: list[ProposedChange] = Field(default=[])

    def extend_changes(self, changes: list[ProposedChange]):
        for change in changes:
            if change.exceptions_seen:
                self.exceptions_seen.append(change)
            else:
                self.proposed_changes.append(change)


class TemplateChangeDetails(PydanticBaseModel):
    resource_id: str
    resource_type: str
    template_path: Union[str, Path]
    # Supports multi-account providers and single-account providers
    proposed_changes: list[Union[AccountChangeDetails, ProposedChange]] = []
    exceptions_seen: list[Union[AccountChangeDetails, ProposedChange]] = Field(
        default=[]
    )

    class Config:
        json_encoders = {PrettyOrderedSet: list}
        extra = Extra.forbid

    @validator("template_path")
    def validate_template_path(cls, v: Union[str, Path]):
        return str(v)

    def extend_changes(
        self, changes: list[Union[AccountChangeDetails, ProposedChange]]
    ):
        for change in changes:
            if change.exceptions_seen:
                if isinstance(change, AccountChangeDetails) and change.proposed_changes:
                    # If there was a partial failure then split the successes into their own change
                    proposed_change = change.copy()
                    proposed_change.exceptions_seen = []
                    self.proposed_changes.append(proposed_change)
                    change.proposed_changes = (
                        []
                    )  # Remove the proposed changes from the original change
                self.exceptions_seen.append(change)
            elif isinstance(change, AccountChangeDetails) and change.proposed_changes:
                self.proposed_changes.append(change)
            elif isinstance(change, ProposedChange):
                self.proposed_changes.append(change)

    def dict(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
    ) -> Dict[str, Any]:  # noqa
        response = self.json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return json.loads(response)


class ProviderChild(PydanticBaseModel):
    """
    Inherited by the provider class to provide a consistent interface for AccessModelMixin

    For AWS, this is the AWS account
    For GCP, this is the GCP project
    For Okta, this is the IDP domain
    """

    iambic_managed: Optional[IambicManaged] = Field(
        IambicManaged.UNDEFINED,
        description="Controls the directionality of iambic changes",
    )

    @property
    def parent_id(self) -> Optional[str]:
        """
        For example, the parent_id of an AWS account is the AWS organization ID
        """
        raise NotImplementedError

    @property
    def preferred_identifier(self) -> str:
        raise NotImplementedError

    @property
    def all_identifiers(self) -> set[str]:
        raise NotImplementedError


class AccessModelMixin:
    @property
    def included_children(self):
        raise NotImplementedError

    def set_included_children(self, value):
        raise NotImplementedError

    @property
    def excluded_children(self):
        raise NotImplementedError

    def set_excluded_children(self, value):
        raise NotImplementedError

    @property
    def included_parents(self):
        raise NotImplementedError

    def set_included_parents(self, value):
        raise NotImplementedError

    @property
    def excluded_parents(self):
        raise NotImplementedError

    def set_excluded_parents(self, value):
        raise NotImplementedError

    def access_model_sort_weight(self):
        # we have to pay the price eo sort it before using the value
        # because the validators are only called during model creation
        # and others have may have mutate the list value
        sorted_included_children = sorted(self.included_children)
        sorted_excluded_children = sorted(self.excluded_children)
        sorted_included_parents = sorted(self.included_parents)
        sorted_excluded_parents = sorted(self.excluded_parents)

        return (
            str(sorted_included_children)
            + str(sorted_excluded_children)
            + str(sorted_included_parents)
            + str(sorted_excluded_parents)
        )


class BaseTemplate(
    BaseModel,
):
    template_type: str
    file_path: Union[str, Path] = Field(..., hidden_from_schema=True)
    owner: Optional[str]
    iambic_managed: Optional[IambicManaged] = Field(
        IambicManaged.UNDEFINED,
        description="Controls the directionality of Iambic changes",
    )

    def dict(
        self,
        *,
        include: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        exclude: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
    ) -> "DictStrAny":  # noqa
        if exclude:
            exclude.add("file_path")
        else:
            exclude = {"file_path"}

        template_dict = self.json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        template_dict = json.loads(template_dict)
        template_dict["template_type"] = self.template_type
        return template_dict

    def get_body(self, exclude_none=True, exclude_unset=True, exclude_defaults=True):
        input_dict = self.dict(
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude={"file_path"},
        )
        sorted_input_dict = sort_dict(input_dict)
        sorted_input_dict = create_commented_map(sorted_input_dict)
        as_yaml = yaml.dump(sorted_input_dict)
        # Force template_type to be at the top of the yaml
        template_type_str = f"template_type: {self.template_type}"
        as_yaml = as_yaml.replace(f"{template_type_str}\n", "")
        as_yaml = as_yaml.replace(f"\n{template_type_str}", "")
        as_yaml = f"{template_type_str}\n{as_yaml}"
        return as_yaml

    def write(self, exclude_none=True, exclude_unset=True, exclude_defaults=True):
        # pay the cost of validating the models once more.
        self.validate_model_afterward()

        as_yaml = self.get_body(exclude_none, exclude_unset, exclude_defaults)
        os.makedirs(os.path.dirname(os.path.expanduser(self.file_path)), exist_ok=True)
        with open(self.file_path, "w") as f:
            f.write(as_yaml)

    def delete(self):
        log.info("Deleting template file", file_path=self.file_path)
        try:
            repo = Repo(self.file_path, search_parent_directories=True)
            # why force=True? Expire could have modified the local contents
            # without force=True, git rm would not be able to remove the file
            repo.index.remove(
                [
                    str(self.file_path)
                ],  # manual cast to str is necessary because git library only accepts str and not FilePath
                working_tree=True,
                force=True,
            )
        except Exception as e:
            log.error(
                "Unable to remove file from local Git repo. Deleting manually",
                error=e,
                file_path=self.file_path,
            )
            os.remove(self.file_path)

    async def apply(self, config: Config) -> TemplateChangeDetails:
        raise NotImplementedError

    @classmethod
    def load(cls, file_path: str):
        return cls(
            file_path=file_path, **transform_comments(yaml.load(open(file_path)))
        )

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"iambic_managed", "file_path", "owner"}


class Variable(PydanticBaseModel):
    key: str
    value: str

    def __getitem__(self, item):
        return getattr(self, item)


class ExpiryModel(IambicPydanticBaseModel):
    expires_at: Optional[Union[str, datetime.datetime, datetime.date]] = Field(
        None, description="The date and time the resource will be/was set to deleted."
    )
    deleted: Optional[bool] = Field(
        False,
        description=(
            "Denotes whether the resource has been removed from AWS."
            "Upon being set to true, the resource will be deleted the next time iambic is ran."
        ),
    )

    class Config:
        json_encoders = {
            datetime.datetime: simplify_dt,
            datetime.date: simplify_dt,
        }
        extra = Extra.forbid

    @validator("expires_at", pre=True)
    def parse_expires_at(cls, value):
        if not value:
            return value
        if isinstance(value, datetime.date) and not isinstance(
            value, datetime.datetime
        ):
            dt = datetime.datetime.combine(value, datetime.datetime.min.time())
            dt = dt.replace(tzinfo=datetime.timezone.utc)
            return dt
        if isinstance(value, datetime.datetime):
            dt = value
            if not dt.tzinfo:
                dt = dt.replace(tzinfo=datetime.timezone.utc)
            return dt
        dt = dateparser.parse(
            value, settings={"TIMEZONE": "UTC", "RETURN_AS_TIMEZONE_AWARE": True}
        )
        return dt

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"expires_at", "deleted"}


class ExecutionMessage(PydanticBaseModel):
    execution_id: str = Field(..., description="A unique identifier for the execution")
    command: Command
    parent_command: Optional[Command]
    provider_type: Optional[str]
    provider_id: Optional[str]
    template_type: Optional[str]
    template_id: Optional[str]
    metadata: Optional[Dict[str, Any]] = None
    templates: Optional[List[str]] = None

    @root_validator
    def check_parent_command(cls, values: dict):
        if not values.get("parent_command"):
            values["parent_command"] = values.get("command")

        return values

    def get_execution_dir(self, as_regex: bool = False) -> str:
        if as_regex:
            as_regex = "**"

        path_params = [self.execution_id]
        if path_param := self.provider_type or as_regex:
            path_params.append(path_param)
        if path_param := self.provider_id or as_regex:
            path_params.append(path_param)
        if path_param := self.template_type or as_regex:
            path_params.append(path_param)
        if path_param := self.template_id or as_regex:
            path_params.append(path_param)
        if path_param := self.metadata or as_regex:
            if path_param != as_regex:
                path_param = md5(
                    json.dumps(path_param, sort_keys=True).encode("utf-8")
                ).hexdigest()
            path_params.append(path_param)

        file_path = os.path.join(get_writable_directory(), ".iambic", *path_params)
        if not as_regex:
            os.makedirs(file_path, exist_ok=True)

        return file_path

    def get_directory(self, *path_dirs) -> str:
        dir_path = os.path.join(self.get_execution_dir(), *path_dirs)
        os.makedirs(dir_path, exist_ok=True)
        return dir_path

    def get_file_path(self, *path_dirs, file_name_and_extension: str) -> str:
        return os.path.join(self.get_directory(*path_dirs), file_name_and_extension)

    async def get_sub_exe_files(
        self,
        *path_dirs,
        file_name_and_extension: str = None,
        flatten_results: bool = False,
    ) -> List[dict]:
        async def _get_file_contents(file_path: str) -> dict:
            async with aiofiles.open(file_path, mode="r") as f:
                file_text = await f.read()
                if file_path.endswith(".yaml") or file_path.endswith(".yml"):
                    return yaml.load(file_text)
                elif file_path.endswith(".json"):
                    return json.loads(file_text)
                else:
                    raise TypeError(
                        "Unsupported file type. Must be one of yaml, yml, or json."
                    )

        matching_files = glob.glob(
            os.path.join(
                self.get_execution_dir(True),
                *path_dirs if path_dirs else "**",
                file_name_and_extension or "**",
            ),
            recursive=True,
        )
        response = list(
            await asyncio.gather(
                *[_get_file_contents(file_path) for file_path in set(matching_files)]
            )
        )
        if flatten_results:
            return list(itertools.chain.from_iterable(response))

        return response


class ExecutionResponse(ExecutionMessage):
    status: ExecutionStatus
    errors: Optional[list[str]]



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.github.iambic_plugin import GithubConfig


async def load(config: GithubConfig) -> GithubConfig:
    import iambic.plugins.v0_1_0.github.github

    iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_USER_NAME = (
        config.commit_message_user_name
    )
    iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_USER_EMAIL = (
        config.commit_message_user_email
    )
    iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_FOR_DETECT = (
        config.commit_message_for_detect
    )
    iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_FOR_IMPORT = (
        config.commit_message_for_import
    )
    iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_FOR_EXPIRE = (
        config.commit_message_for_expire
    )
    iambic.plugins.v0_1_0.github.github.COMMIT_MESSAGE_FOR_GIT_APPLY_ABSOLUTE_TIME = (
        config.commit_message_for_git_apply
    )

    return config


# This is not managing Github cloud resources today.
# It exists to comply with plugin interface. Github Plugin
# today provides GithubApp integrations for iambic templates
async def import_github_resources(
    config: GithubConfig, base_output_dir: str, messages: list = None
):
    tasks = []
    await asyncio.gather(*tasks)



from __future__ import annotations

from pydantic import BaseModel

from iambic.core.iambic_plugin import ProviderPlugin
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.example.handlers import import_example_resources, load
from iambic.plugins.v0_1_0.example.local_database.models import (
    ExampleLocalDatabaseTemplate,
)
from iambic.plugins.v0_1_0.example.local_file.models import (
    ExampleLocalFileMultiAccountTemplate,
    ExampleLocalFileTemplate,
)


class ExampleConfig(BaseModel):
    pass


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="example",
    version=PLUGIN_VERSION,
    provider_config=ExampleConfig,
    requires_secret=True,
    async_import_callable=import_example_resources,
    async_load_callable=load,
    templates=[
        ExampleLocalFileTemplate,
        ExampleLocalFileMultiAccountTemplate,
        ExampleLocalDatabaseTemplate,
    ],
)



from __future__ import annotations

import pathlib

from jinja2 import Environment, FileSystemLoader

from iambic.output.filters import (
    rich_format,
    rich_text,
    rich_text_table,
    rich_tree_exception,
    rich_tree_summary,
)


def get_template_env():
    my_path = pathlib.Path(__file__).parent.absolute()
    env = Environment(loader=FileSystemLoader(my_path / "templates"))
    env.filters["rich_format"] = rich_format
    env.filters["rich_text"] = rich_text
    env.filters["rich_text_table"] = rich_text_table
    env.filters["rich_tree_summary"] = rich_tree_summary
    env.filters["rich_tree_exception"] = rich_tree_exception
    return env



from __future__ import annotations

import asyncio
import os
from typing import TYPE_CHECKING

from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template as common_create_or_update_template,
)
from iambic.core.template_generation import (
    delete_orphaned_templates,
    get_existing_template_map,
)
from iambic.plugins.v0_1_0.google_workspace.group.models import (
    GOOGLE_GROUP_TEMPLATE_TYPE,
    GoogleWorkspaceGroupTemplate,
)
from iambic.plugins.v0_1_0.google_workspace.group.utils import list_groups

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
        GoogleProject,
        GoogleWorkspaceConfig,
    )


def get_resource_dir_args(domain: str) -> list:
    return ["group", domain]


def get_response_dir(
    exe_message: ExecutionMessage, google_project: GoogleProject, domain: str
) -> str:
    dir_args = get_resource_dir_args(domain)
    dir_args.append("templates")
    if exe_message.provider_id:
        return exe_message.get_directory(*dir_args)
    else:
        return exe_message.get_directory(google_project.project_id, *dir_args)


def get_group_dir(
    base_dir: str,
    domain: str,
) -> str:
    return str(
        os.path.join(
            base_dir, "resources", "google_workspace", *get_resource_dir_args(domain)
        )
    )


def get_templated_resource_file_path(
    resource_dir: str,
    resource_email: str,
) -> str:
    unwanted_chars = ["}}_", "}}", ".", "-", " "]
    resource_name = resource_email.split("@")[0].replace("{{", "").lower()
    for unwanted_char in unwanted_chars:
        resource_name = resource_name.replace(unwanted_char, "_")
    return str(os.path.join(resource_dir, f"{resource_name}.yaml"))


async def generate_domain_group_resource_files(
    exe_message: ExecutionMessage, project: GoogleProject, domain: str
):
    account_group_response_dir = get_response_dir(exe_message, project, domain)

    log.info("Caching Google groups.", project_id=project.project_id, domain=domain)

    groups = await list_groups(domain, project)
    for group in groups:
        group.file_path = os.path.join(
            account_group_response_dir, f"{group.properties.email.split('@')[0]}.yaml"
        )
        group.write()

    log.info(
        "Finished caching Google groups.",
        project_id=project.project_id,
        domain=domain,
        group_count=len(groups),
    )


async def update_or_create_group_template(
    discovered_group_template: GoogleWorkspaceGroupTemplate,
    existing_template_map: dict,
    group_dir: str,
) -> GoogleWorkspaceGroupTemplate:
    discovered_group_template.file_path = get_templated_resource_file_path(
        group_dir,
        discovered_group_template.properties.email,
    )

    return common_create_or_update_template(
        discovered_group_template.file_path,
        existing_template_map,
        discovered_group_template.resource_id,
        GoogleWorkspaceGroupTemplate,
        {},
        discovered_group_template.properties,
        [],
    )


async def collect_project_groups(
    exe_message: ExecutionMessage, config: GoogleWorkspaceConfig
):
    assert exe_message.provider_id
    project = config.get_workspace(exe_message.provider_id)
    log.info("Beginning to retrieve Google groups.", project=project.project_id)

    await asyncio.gather(
        *[
            generate_domain_group_resource_files(exe_message, project, subject.domain)
            for subject in project.subjects
        ]
    )

    log.info("Finished retrieving Google group details", project=project.project_id)


async def generate_group_templates(
    exe_message: ExecutionMessage,
    config: GoogleWorkspaceConfig,
    output_dir: str,
    detect_messages: list = None,
):
    """List all groups in the domain, along with members and
    settings"""

    base_path = os.path.expanduser(output_dir)
    existing_template_map = await get_existing_template_map(
        base_path, GOOGLE_GROUP_TEMPLATE_TYPE
    )
    all_resource_ids = set()

    log.info("Updating and creating Google group templates.")

    for workspace in config.workspaces:
        for subject in workspace.subjects:
            domain = subject.domain
            group_dir = get_group_dir(base_path, domain)

            groups = await exe_message.get_sub_exe_files(
                *get_resource_dir_args(domain),
                "templates",
                file_name_and_extension="**.yaml",
            )
            # Update or create templates
            for group in groups:
                group = GoogleWorkspaceGroupTemplate(file_path="unset", **group)
                group.file_path = group.default_file_path
                resource_template = await update_or_create_group_template(
                    group, existing_template_map, group_dir
                )
                if not resource_template:
                    # Template not updated. Most likely because it's an `enforced` template.
                    continue
                all_resource_ids.add(resource_template.resource_id)

    # Delete templates that no longer exist
    delete_orphaned_templates(list(existing_template_map.values()), all_resource_ids)

    log.info("Finish updating and creating Google group templates.")



from __future__ import annotations

from pydantic import BaseModel as PydanticBaseModel


class PermissionSetMessageDetails(PydanticBaseModel):
    account_id: str
    instance_arn: str
    permission_set_arn: str


class ManagedPolicyMessageDetails(PydanticBaseModel):
    account_id: str
    policy_path: str
    policy_name: str
    delete: bool


class RoleMessageDetails(PydanticBaseModel):
    account_id: str
    role_name: str
    delete: bool


class GroupMessageDetails(PydanticBaseModel):
    account_id: str
    group_name: str
    delete: bool


class UserMessageDetails(PydanticBaseModel):
    account_id: str
    user_name: str
    delete: bool



from __future__ import annotations

import asyncio
from enum import Enum
from typing import TYPE_CHECKING

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseTemplate,
    ExpiryModel,
    TemplateChangeDetails,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import (
        GoogleProject,
        GoogleWorkspaceConfig,
    )


class WhoCanInvite(Enum):
    ALL_MANAGERS_CAN_INVITE = "ALL_MANAGERS_CAN_INVITE"
    ALL_MEMBERS_CAN_INVITE = "ALL_MEMBERS_CAN_INVITE"


class WhoCanJoin(Enum):
    ALL_IN_DOMAIN_CAN_JOIN = "ALL_IN_DOMAIN_CAN_JOIN"
    ANYONE_CAN_JOIN = "ANYONE_CAN_JOIN"
    CAN_REQUEST_TO_JOIN = "CAN_REQUEST_TO_JOIN"


class WhoCanPostMessage(Enum):
    ALL_IN_DOMAIN_CAN_POST = "ALL_IN_DOMAIN_CAN_POST"
    ALL_MANAGERS_CAN_POST = "ALL_MANAGERS_CAN_POST"
    ALL_MEMBERS_CAN_POST = "ALL_MEMBERS_CAN_POST"
    ANYONE_CAN_POST = "ANYONE_CAN_POST"
    NONE_CAN_POST = "NONE_CAN_POST"


class WhoCanViewGroup(Enum):
    ALL_IN_DOMAIN_CAN_VIEW = "ALL_IN_DOMAIN_CAN_VIEW"
    ALL_MANAGERS_CAN_VIEW = "ALL_MANAGERS_CAN_VIEW"
    ALL_MEMBERS_CAN_VIEW = "ALL_MEMBERS_CAN_VIEW"
    ANYONE_CAN_VIEW = "ANYONE_CAN_VIEW"


class WhoCanViewMembership(Enum):
    ALL_IN_DOMAIN_CAN_VIEW = "ALL_IN_DOMAIN_CAN_VIEW"
    ALL_MANAGERS_CAN_VIEW = "ALL_MANAGERS_CAN_VIEW"
    ALL_MEMBERS_CAN_VIEW = "ALL_MEMBERS_CAN_VIEW"
    ANYONE_CAN_VIEW = "ANYONE_CAN_VIEW"


class GroupMemberRole(Enum):
    OWNER = "OWNER"
    MANAGER = "MANAGER"
    MEMBER = "MEMBER"


class GroupMemberSubscription(Enum):
    EACH_EMAIL = "EACH_EMAIL"
    DIGEST = "DIGEST"
    ABRIDGED = "ABRIDGED"
    NO_EMAIL = "NO_EMAIL"


class Posting(Enum):
    ALLOWED = "ALLOWED"
    NOT_ALLOWED = "NOT_ALLOWED"
    MODERATED = "MODERATED"


class GroupMemberType(Enum):
    USER = "USER"
    GROUP = "GROUP"
    EXTERNAL = "EXTERNAL"


class GroupMemberStatus(Enum):
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    PENDING = "PENDING"
    UNDEFINED = "UNDEFINED"


class GoogleTemplate(BaseTemplate, ExpiryModel):
    async def _apply_to_account(
        self,
        google_project: GoogleProject,
    ) -> AccountChangeDetails:
        raise NotImplementedError

    async def apply(self, config: GoogleWorkspaceConfig) -> TemplateChangeDetails:
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.properties.email,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        log_params = dict(
            resource_type=self.resource_type, resource_id=self.resource_id
        )
        for account in config.workspaces:
            match = False
            for subject in account.subjects:
                if self.properties.domain == subject.domain:
                    match = True
            if not match:
                continue
            # if evaluate_on_google_account(self, account):
            if ctx.execute:
                log_str = "Applying changes to resource."
            else:
                log_str = "Detecting changes for resource."
            log.info(log_str, **log_params)
            tasks.append(self._apply_to_account(account))

        account_changes = await asyncio.gather(*tasks)
        template_changes.proposed_changes = [
            account_change
            for account_change in account_changes
            if any(account_change.proposed_changes)
        ]

        proposed_changes = [x for x in account_changes if x.proposed_changes]

        if proposed_changes and ctx.execute:
            log.info(
                "Successfully applied all or some resource changes to all Google projects. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        elif proposed_changes:
            log.info(
                "Successfully detected all or some required resource changes on all Google projects. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        else:
            log.debug("No changes detected for resource on any account.", **log_params)

        return template_changes

    @property
    def resource_id(self) -> str:
        return self.properties.email

    @property
    def resource_type(self) -> str:
        return "google:group"



from __future__ import annotations

import os
import re
from io import StringIO
from typing import TYPE_CHECKING, Optional

from deepdiff import DeepDiff
from git import Repo
from git.exc import GitCommandError
from pydantic import BaseModel as PydanticBaseModel

from iambic.config.templates import TEMPLATES
from iambic.core.logger import log
from iambic.core.parser import load_templates
from iambic.core.utils import NOQ_TEMPLATE_REGEX, file_regex_search, yaml

if TYPE_CHECKING:
    from iambic.config.dynamic_config import Config


class GitDiff(PydanticBaseModel):
    path: str
    content: Optional[str] = None
    is_deleted: Optional[bool] = False


def get_origin_head(repo: Repo) -> bool:
    default_branch = [x for x in repo.remotes.origin.refs if x.name == "origin/HEAD"]
    if any(default_branch):
        return default_branch[0].name.split("/")[-1]
    else:
        raise ValueError(
            "Unable to determine the default branch for the repo 'origin' remote"
        )


async def clone_git_repos(config, repo_dir: str) -> dict[str, Repo]:
    # TODO: Formalize the model for secrets
    repos = {}
    for repository in config.secrets.get("git", {}).get("repositories", []):
        repo_name = repository["name"]
        git_uri = repository["uri"]
        repo_path = os.path.join(repo_dir, repo_name)
        try:
            repo = Repo.clone_from(git_uri, repo_path)
            repos[repo_name] = repo
        except GitCommandError as e:
            if "already exists and is not an empty directory" not in e.stderr:
                raise
            repo = Repo(repo_path)
            for remote in repo.remotes:
                remote.fetch()
            repo.git.pull()
            repos[repo_name] = repo
    return repos


def clone_git_repo(repo_url: str, repo_path: str, remote_branch_name: str):
    repo = Repo.clone_from(repo_url, repo_path, branch=remote_branch_name)
    return repo


def get_remote_default_branch(repo: Repo, remote_name: str = "origin"):
    # This is relying on `git remote show origin`
    # includes information  HEAD branch: THE_ACTUAL_BRANCH_NAME
    #
    remote_info_lines = repo.git.remote("show", remote_name).split("\n")
    default_branch = ""
    for line in remote_info_lines:
        if "HEAD branch" in line:
            default_branch = line.split(":")[1].strip()
            break
    if default_branch == "":
        default_branch = "main"
    return default_branch


async def retrieve_git_changes(
    repo_dir: str,
    allow_dirty: bool = False,
    from_sha=None,
    to_sha=None,
) -> dict[str, list[GitDiff]]:
    repo = Repo(repo_dir)
    if repo.is_dirty():
        log.error(
            "Template git repo is dirty, and `allow_dirty` is not enabled. "
            "Refusing to proceed",
            file_path=repo_dir,
        )

    from_sha_obj = None
    to_sha_obj = None

    if from_sha is None:
        # Fetch latest
        for remote in repo.remotes:
            remote.fetch()
        # Comparing against default_branch
        remote_name = "origin"
        default_branch = get_remote_default_branch(repo, remote_name)
        commit_origin_main = repo.commit(f"{remote_name}/{default_branch}")
        # TODO: We should consider if the default branch is named other than `main`
        from_sha_obj = commit_origin_main
    else:
        from_sha_obj = repo.commit(from_sha)
    if to_sha is None:
        # Last commit of the current branch
        commit_feature = repo.head.commit.tree
        to_sha_obj = commit_feature
    else:
        to_sha_obj = repo.commit(to_sha)

    diff_index = from_sha_obj.diff(to_sha_obj)
    files = {
        "new_files": [],
        "deleted_files": [],
        "modified_files": [],
    }

    # Collect all new files
    for file_obj in diff_index.iter_change_type("A"):
        if (path := str(os.path.join(repo_dir, file_obj.b_path))).endswith(
            ".yaml"
        ) and (await file_regex_search(path, NOQ_TEMPLATE_REGEX)):
            file = GitDiff(path=str(os.path.join(repo_dir, path)))
            files["new_files"].append(file)

    # Collect all deleted files
    if (
        False
    ):  # EN-1635 Disable file deletion cleanup because we don't handle backward compatible and worry about accidental removal
        for file_obj in diff_index.iter_change_type("D"):
            if (path := file_obj.b_path).endswith(".yaml"):
                file = GitDiff(
                    path=str(os.path.join(repo_dir, path)),
                    content=file_obj.a_blob.data_stream.read().decode("utf-8"),
                    is_deleted=True,
                )
                if re.search(NOQ_TEMPLATE_REGEX, file.content):
                    files["deleted_files"].append(file)

    # Collect all modified files
    for file_obj in diff_index.iter_change_type("M"):
        if (path := str(os.path.join(repo_dir, file_obj.b_path))).endswith(
            ".yaml"
        ) and (await file_regex_search(path, NOQ_TEMPLATE_REGEX)):
            if (
                main_path := str(os.path.join(repo_dir, file_obj.a_path))
            ) != path:  # File was renamed
                deleted_file = GitDiff(
                    path=str(os.path.join(repo_dir, main_path)),
                    content=file_obj.a_blob.data_stream.read().decode("utf-8"),
                    is_deleted=True,
                )

                if re.search(NOQ_TEMPLATE_REGEX, deleted_file.content):
                    template_dict = yaml.load(open(path))
                    main_template_dict = yaml.load(StringIO(deleted_file.content))
                    if not DeepDiff(
                        template_dict,
                        main_template_dict,
                        ignore_order=True,
                        report_repetition=True,
                    ):
                        continue  # Just renamed but no file changes

                    template_cls = TEMPLATES.template_map[
                        main_template_dict["template_type"]
                    ]
                    main_template = template_cls(
                        file_path=deleted_file.path, **main_template_dict
                    )
                    template = template_cls(file_path=path, **template_dict)
                    if main_template.resource_id != template.resource_id:
                        files["deleted_files"].append(deleted_file)
                        files["new_files"].append(GitDiff(path=path))
                        continue

            file = GitDiff(
                path=str(os.path.join(repo_dir, path)),
                content=file_obj.a_blob.data_stream.read().decode("utf-8"),
            )
            files["modified_files"].append(file)

    return files


def create_templates_for_deleted_files(deleted_files: list[GitDiff]) -> list:
    """
    Create a class instance of the deleted file content with its template type
    If it wasn't deleted, set it to deleted
    Add that instance to templates
    """
    templates = []
    for git_diff in deleted_files:
        template_dict = yaml.load(StringIO(git_diff.content))
        template_cls = TEMPLATES.template_map[template_dict["template_type"]]
        template = template_cls(file_path=git_diff.path, **template_dict)
        if template.deleted is True:
            continue
        template.deleted = True
        log.info("Template marked as deleted", file_path=git_diff.path)
        templates.append(template)

    return templates


def create_templates_for_modified_files(
    config: Config, modified_files: list[GitDiff]
) -> list:
    """
    Create a class instance of the original file content and the new file content with its template type
    Check for aws_accounts that were removed from included_accounts or added to excluded_accounts
    Update the template to be applied to delete the role from the aws_accounts that hit on the above statement
    """
    templates = []
    for git_diff in modified_files:
        main_template_dict = yaml.load(StringIO(git_diff.content))
        template_type_string = main_template_dict["template_type"]
        template_cls = TEMPLATES.template_map.get(template_type_string, None)

        if template_cls is None:
            # well the case is the previous version is an unknown config type now.
            # this is typically the config file
            log_params = {"template_type": template_type_string}
            log.warning(
                "template_type is not registered among template_map", **log_params
            )
            continue

        main_template = template_cls(file_path=git_diff.path, **main_template_dict)

        # template_dict = yaml.load(open(git_diff.path))
        # template = template_cls(file_path=git_diff.path, **template_dict)
        template = load_templates([git_diff.path])[0]

        # EN-1634 dealing with providers that have no concept of multi-accounts
        # a hack to just ignore template that does not have included_accounts attribute
        if getattr(main_template, "included_accounts", None) is None:
            templates.append(template)
            # The rest of the account inclusion/exclusion logic does not apply
            # plugin that does not have concept of included_accounts
            continue

        deleted_included_accounts = []
        # deleted_exclude_accounts are aws_accounts that are included in the current commit so can't be deleted
        deleted_exclude_accounts = [*template.included_accounts]
        deleted_exclude_accounts_str = "\n".join(deleted_exclude_accounts)

        # Catch aws_accounts that were in included aws_accounts but have been removed
        if "*" not in deleted_exclude_accounts:
            if "*" in main_template.included_accounts:
                """
                Catch aws_accounts that were implicitly removed from included_accounts.
                Example:
                    main branch included_accounts:
                        - *
                    current commit included_accounts:
                        - staging
                        - dev

                If config.aws.accounts included prod, staging, and dev this will catch that prod is no longer included.
                    This means marking prod for deletion as it has been implicitly deleted.
                """
                for aws_account in config.aws.accounts:
                    account_regex = (
                        rf"({aws_account.account_id}|{aws_account.account_name})"
                    )
                    if re.search(
                        re.escape(account_regex), deleted_exclude_accounts_str
                    ):
                        log.debug(
                            "Resource on account not marked deletion.",
                            account=account_regex,
                            template=git_diff.path,
                        )
                        continue

                    log.info(
                        "Marking resource for deletion on account.",
                        reason="Implicitly removed from included_accounts",
                        account=account_regex,
                        template=git_diff.path,
                    )
                    deleted_included_accounts.append(account_regex)
                    template.included_accounts.append(account_regex)
            else:
                """
                Catch aws_accounts that were explicitly removed from included_accounts.
                Example:
                    main branch included_accounts:
                        - prod
                        - staging
                        - dev
                    current commit included_accounts:
                        - staging
                        - dev

                This means marking prod for deletion as it has been implicitly deleted.
                """
                for account in main_template.included_accounts:
                    if re.search(re.escape(account), deleted_exclude_accounts_str):
                        log.debug(
                            "Resource on account not marked deletion.",
                            account=account,
                            template=git_diff.path,
                        )
                        continue

                    log.info(
                        "Marking resource for deletion on account.",
                        reason="Explicitly removed from included_accounts.",
                        account=account,
                        template=git_diff.path,
                    )
                    deleted_included_accounts.append(account)
                    template.included_accounts.append(account)

        main_template_included_accounts_str = "\n".join(main_template.included_accounts)
        main_template_excluded_accounts_str = (
            "\n".join(main_template.excluded_accounts)
            if main_template.excluded_accounts
            else None
        )
        template_excluded_accounts = []
        """
        Catch aws_accounts that have been implicitly excluded.
        Example:
            included_accounts:
                - *
            main branch excluded_accounts: []
            current commit excluded_accounts:
                - prod

        If config.aws.accounts included prod, staging, and dev this will catch that prod is no longer included.
            This means marking prod for deletion as it has been implicitly deleted.
        """
        for account in template.excluded_accounts:
            if main_template_excluded_accounts_str and re.search(
                re.escape(account), main_template_excluded_accounts_str
            ):
                # The account was already excluded so add it to the template_excluded_accounts
                log.debug(
                    "Resource already excluded on account.",
                    account=account,
                    template=git_diff.path,
                )
                template_excluded_accounts.append(account)
            elif (
                re.search(re.escape(account), main_template_included_accounts_str)
                or "*" in main_template.included_accounts
            ):
                # The account was previously included so mark it for deletion
                log.info(
                    "Marking resource for deletion on account.",
                    reason="Account added to excluded_accounts for resource.",
                    account=account,
                    template=git_diff.path,
                )
                deleted_included_accounts.append(account)
                template.included_accounts.append(account)
            else:
                # The account wasn't included or excluded before this so add it back to template_excluded_accounts
                log.debug(
                    "Newly excluded account.", account=account, template=git_diff.path
                )
                template_excluded_accounts.append(account)

        template.excluded_accounts = template_excluded_accounts

        if deleted_included_accounts and template.deleted is not True:
            template.deleted = True

        templates.append(template)

    return templates



from __future__ import annotations

import asyncio
import re
from itertools import chain
from typing import TYPE_CHECKING, Callable, List, Optional, Union

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import (
    AccessModelMixin,
    AccountChangeDetails,
    BaseModel,
    ProposedChange,
    ProposedChangeType,
    TemplateChangeDetails,
)
from iambic.core.utils import aio_wrapper, evaluate_on_provider, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.iam.policy.models import PolicyStatement
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.utils import (
    apply_account_assignments,
    apply_permission_set_aws_managed_policies,
    apply_permission_set_customer_managed_policies,
    apply_permission_set_inline_policy,
    apply_permission_set_permission_boundary,
    apply_permission_set_tags,
    delete_permission_set,
    enrich_permission_set_details,
    get_permission_set_users_and_groups_as_access_rules,
)
from iambic.plugins.v0_1_0.aws.models import (
    AccessModel,
    AWSAccount,
    AWSTemplate,
    Description,
    ExpiryModel,
    Tag,
)
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE = (
    "NOQ::AWS::IdentityCenter::PermissionSet"
)

# TODO: Add true support for defining multiple orgs with IdentityCenter rules


class PermissionSetAccess(AccessModel, ExpiryModel):
    users: list[str] = Field(
        [],
        description="List of users who can access the role",
    )
    groups: list[str] = Field(
        [],
        description="List of groups. Users in one or more of the groups can access the role",
    )

    @property
    def resource_type(self):
        return "aws:identity_center:permission_set_access"

    @property
    def resource_id(self):
        return ""


class AWSIdentityCenterInstance(BaseModel):
    arn: str
    region: str
    access_portal_url: str
    identity_store_id: str


class CustomerManagedPolicyReference(BaseModel, ExpiryModel):
    path: Optional[str] = Field(
        "/",
        description="The path to the IAM policy that you have configured in each account where you want to deploy your permission set. The default is /. For more information, see Friendly names and paths in the IAM User Guide.",
    )
    name: str

    @property
    def resource_type(self) -> str:
        return "aws:iam:managed_policy"

    @property
    def resource_id(self) -> str:
        return f"{self.path}{self.name}"


class ManagedPolicyArn(BaseModel, ExpiryModel):
    arn: str

    @property
    def resource_type(self) -> str:
        return "aws:iam:managed_policy"

    @property
    def resource_id(self) -> str:
        return self.arn


class PermissionBoundary(BaseModel, ExpiryModel):
    customer_managed_policy_reference: Optional[CustomerManagedPolicyReference]
    managed_policy_arn: Optional[str]

    @property
    def resource_type(self):
        return "aws:identity_center:permission_boundary"

    @property
    def resource_id(self):
        return self.customer_managed_policy_reference.name or self.managed_policy_arn


class SessionDuration(BaseModel):
    session_duration: str


class InlinePolicy(BaseModel, ExpiryModel):
    version: Optional[str] = None
    statement: Optional[List[PolicyStatement]] = Field(
        None,
        description="List of policy statements",
    )

    @property
    def resource_type(self) -> str:
        return "aws:identity_center:inline_policy"

    @property
    def resource_id(self) -> str:
        return str(self.statement)


class PermissionSetProperties(BaseModel):
    name: str
    description: Optional[Union[str, list[Description]]] = Field(
        None,
        description="Description of the permission set",
    )
    relay_state: Optional[str] = None
    session_duration: Optional[Union[str, list[SessionDuration]]] = None
    permissions_boundary: Optional[PermissionBoundary] = None
    inline_policy: Optional[InlinePolicy] = None
    customer_managed_policy_references: Optional[
        list[CustomerManagedPolicyReference]
    ] = []
    managed_policies: Optional[list[ManagedPolicyArn]] = []
    tags: Optional[list[Tag]] = []

    @property
    def resource_type(self) -> str:
        return "aws:identity_center:permission_set"

    @property
    def resource_id(self) -> str:
        return self.name

    @classmethod
    def sort_func(cls, attribute_name: str) -> Callable:
        def _sort_func(obj):
            return f"{getattr(obj, attribute_name)}!{obj.access_model_sort_weight()}"

        return _sort_func

    @validator("description")
    def validate_description(cls, v: Union[str, list[Description]]):
        # validation portion
        if isinstance(v, str) and not (1 <= len(v) <= 700):
            raise ValueError(
                f"description must be between 1 and 700 characters: given {v}"
            )
        if isinstance(v, list):
            for description in v:
                description: Description
                if not (1 <= len(description.description) <= 700):
                    raise ValueError(
                        f"description must be between 1 and 700 characters: given {description.description}"
                    )

        # sorting portion
        if not isinstance(v, list):
            return v

        return sorted(v, key=lambda d: d.access_model_sort_weight())

    @validator("managed_policies")
    def sort_managed_policy_refs(cls, v: list[ManagedPolicyArn]):
        sorted_v = sorted(v, key=lambda o: o.resource_id)
        return sorted_v

    @validator("customer_managed_policy_references")
    def sort_customer_managed_policy_references(
        cls, v: list[CustomerManagedPolicyReference]
    ):
        sorted_v = sorted(v, key=lambda o: o.resource_id)
        return sorted_v

    @validator("tags")
    def sort_tags(cls, v: list[Tag]):
        sorted_v = sorted(v, key=cls.sort_func("key"))
        return sorted_v


class AwsIdentityCenterPermissionSetTemplate(
    AccessModelMixin, AWSTemplate, ExpiryModel
):
    template_type: str = AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE
    owner: Optional[str] = Field(None, description="Owner of the permission set")
    properties: PermissionSetProperties
    access_rules: Optional[list[PermissionSetAccess]] = []
    included_orgs: list[str] = Field(
        ["*"],
        description=(
            "A list of AWS organization ids this statement applies to. "
            "Org ids can be represented as a regex and string"
        ),
    )
    excluded_orgs: Optional[list[str]] = Field(
        [],
        description=(
            "A list of AWS organization ids this statement explicitly does not apply to. "
            "Org ids can be represented as a regex and string"
        ),
    )

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"access_rules"}

    @validator("access_rules")
    def sort_access_rules(cls, v: list[PermissionSetAccess]):
        sorted_v = sorted(v, key=lambda o: o.access_model_sort_weight())
        return sorted_v

    async def _access_rules_for_account(  # noqa: C901
        self,
        aws_account: AWSAccount,
        account_id: str,
        account_name: str,
        reverse_user_map: dict[str, str],
        reverse_group_map: dict[str, str],
    ) -> dict:
        """
        Builds a list of access rules for a given account.

        return: {
                "account_id": account_id,
                "user": list(user_ids),
                "group": list(group_ids),
            }
        """

        user_assignments = set()
        group_assignments = set()

        for rule in self.access_rules:
            rule_hit = None

            if rule.deleted:
                continue

            # If the account's org is excluded or not included, skip
            if aws_account.org_id in rule.excluded_orgs:
                continue
            elif "*" not in rule.included_orgs and not any(
                org_id == aws_account.org_id for org_id in rule.included_orgs
            ):
                continue

            if account_name:
                account_reprs = [account_id, account_name.lower()]
            else:
                account_reprs = [account_id]

            # Check against the ways an account can be represented
            # Compare against excluded accounts
            # If it hits on an excluded account rule, skip
            for account_repr in account_reprs:
                for resource_account in rule.excluded_accounts:
                    try:
                        is_hit = await aio_wrapper(
                            re.match, resource_account.lower(), account_repr
                        )
                    except Exception:
                        # Catch accounts with a name that is not a valid regex
                        is_hit = bool(resource_account.lower() == account_repr)

                    if is_hit:
                        rule_hit = False
                        break

                if rule_hit is False:
                    break

            if rule_hit is False:
                continue

            if any(
                resource_account == "*" for resource_account in rule.included_accounts
            ):
                rule_hit = True
            else:
                # Check against the ways an account can be represented
                # Compare against included accounts
                # If it hits on an included account rule
                #   Stop the check and add the users and groups on the rule
                for account_repr in account_reprs:
                    for resource_account in rule.included_accounts:
                        try:
                            is_hit = await aio_wrapper(
                                re.match, resource_account.lower(), account_repr
                            )
                        except Exception:
                            # Catch accounts with a name that is not a valid regex
                            is_hit = bool(resource_account.lower() == account_repr)

                        if is_hit:
                            rule_hit = True
                            break

                    if rule_hit:
                        break

            if rule_hit:
                for rule_user in rule.users:
                    if rule_user == "*":
                        user_assignments.update(reverse_user_map.values())
                    elif user_hit := reverse_user_map.get(rule_user):
                        user_assignments.add(user_hit)

                for rule_group in rule.groups:
                    if rule_group == "*":
                        group_assignments.update(reverse_group_map.values())
                    elif group_hit := reverse_group_map.get(rule_group):
                        group_assignments.add(group_hit)

                if len(group_assignments) == len(reverse_group_map) and len(
                    user_assignments
                ) == len(reverse_user_map):
                    break

        if user_assignments or group_assignments:
            return {
                "account_id": account_id,
                "user": list(user_assignments),
                "group": list(group_assignments),
            }

    async def _verbose_access_rules(self, aws_account: AWSAccount) -> list[dict]:
        """
        Generates the list of access rules across all accounts for the org.
        Fans out calls to _access_rules_for_account and formats the results.
        """
        response = []
        reverse_user_map = {
            details["UserName"]: user_id
            for user_id, details in aws_account.identity_center_details.user_map.items()
        }
        reverse_group_map = {
            details["DisplayName"]: group_id
            for group_id, details in aws_account.identity_center_details.group_map.items()
        }

        account_assignments = await asyncio.gather(
            *[
                self._access_rules_for_account(
                    aws_account,
                    account_id,
                    account_name,
                    reverse_user_map,
                    reverse_group_map,
                )
                for account_id, account_name in aws_account.identity_center_details.org_account_map.items()
            ]
        )

        for account_assignment in account_assignments:
            if not account_assignment:
                continue

            for assignment_type in ["user", "group"]:
                resource_type = assignment_type.upper()
                for assignment in account_assignment[assignment_type]:
                    a_account_id = account_assignment["account_id"]
                    response.append(
                        {
                            "account_id": a_account_id,
                            "resource_id": assignment,
                            "resource_type": resource_type,
                            "resource_name": aws_account.identity_center_details.get_resource_name(
                                resource_type, assignment
                            ),
                            "account_name": f"{a_account_id} ({aws_account.identity_center_details.org_account_map[a_account_id]})",
                        }
                    )

        return response

    async def _apply_to_account(  # noqa: C901
        self, aws_account: AWSAccount
    ) -> AccountChangeDetails:
        """Apply the permission set to the given AWS account

        :param aws_account:
        :return:
        """

        identity_center_client = await aws_account.get_boto3_client(
            "sso-admin", region_name=aws_account.identity_center_details.region_name
        )
        instance_arn = aws_account.identity_center_details.instance_arn
        permission_set_arn = None
        # Marking for deletion. This shouldn't be done on the fly.
        # self = await remove_expired_resources(
        #     self, self.resource_type, self.resource_id
        # )
        template_permission_set = self.apply_resource_dict(aws_account)
        name = template_permission_set["Name"]
        template_account_assignments = await self._verbose_access_rules(aws_account)
        account_change_details = AccountChangeDetails(
            org_id=aws_account.org_id,
            account=str(aws_account),
            resource_id=name,
            resource_type=self.resource_type,
            new_value=dict(
                AccountAssignment=template_account_assignments,
                **template_permission_set,
            ),
            proposed_changes=[],
            exceptions_seen=[],
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_id=name,
            account=str(aws_account),
            org_id=aws_account.org_id,
        )

        current_account_assignments = {}
        current_permission_set = (
            aws_account.identity_center_details.permission_set_map.get(name, {})
        )
        if current_permission_set:
            exclude_keys = ["CreatedDate", "PermissionSetArn"]
            permission_set_arn = current_permission_set["PermissionSetArn"]
            current_permission_set = await enrich_permission_set_details(
                identity_center_client, instance_arn, current_permission_set
            )
            if managed_policies := current_permission_set.pop(
                "AttachedManagedPolicies", None
            ):
                current_permission_set["ManagedPolicies"] = managed_policies

            current_account_assignments = (
                await get_permission_set_users_and_groups_as_access_rules(
                    identity_center_client,
                    instance_arn,
                    permission_set_arn,
                    aws_account.identity_center_details.user_map,
                    aws_account.identity_center_details.group_map,
                    aws_account.identity_center_details.org_account_map,
                )
            )
            account_change_details.current_value = {
                k: v for k, v in current_permission_set.items() if k not in exclude_keys
            }  # Create a new dict
            account_change_details.current_value[
                "AccountAssignments"
            ] = current_account_assignments

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                account_change_details.new_value = {}
                return account_change_details

        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        if isinstance(deleted, list):
            deleted = deleted[0].deleted

        if deleted:
            if current_permission_set:
                account_change_details.new_value = None
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.DELETE,
                        resource_id=name,
                        resource_type=self.resource_type,
                    )
                ]
                log_str = "Active resource found with deleted=false."
                if ctx.execute:
                    log_str = f"{log_str} Deleting resource..."
                log.debug(log_str, **log_params)

                if ctx.execute:
                    apply_awaitable = delete_permission_set(
                        identity_center_client,
                        instance_arn,
                        permission_set_arn,
                        current_permission_set,
                        current_account_assignments,
                        log_params,
                    )
                    proposed_changes = await plugin_apply_wrapper(
                        apply_awaitable, proposed_changes
                    )

                account_change_details.extend_changes(proposed_changes)

            return account_change_details

        permission_set_exists = bool(current_permission_set)
        tasks = []

        if permission_set_exists:
            tasks.append(
                apply_permission_set_tags(
                    identity_center_client,
                    instance_arn,
                    permission_set_arn,
                    template_permission_set.get("Tags", []),
                    current_permission_set.get("Tags", []),
                    log_params,
                )
            )

            supported_update_keys = ["Description", "SessionDuration", "RelayState"]
            update_resource_log_params = {**log_params}
            update_resource_params = {}
            for k in supported_update_keys:
                if template_permission_set.get(
                    k
                ) is not None and template_permission_set.get(
                    k
                ) != current_permission_set.get(
                    k
                ):
                    update_resource_log_params[k] = dict(
                        old_value=current_permission_set.get(k),
                        new_value=template_permission_set.get(k),
                    )
                    update_resource_params[k] = template_permission_set.get(k)

            if update_resource_params:
                log_str = "Out of date resource found."
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.UPDATE,
                        resource_id=name,
                        resource_type=self.resource_type,
                    )
                ]
                if ctx.execute:
                    log.info(
                        f"{log_str} Updating resource...",
                        **update_resource_log_params,
                    )
                    apply_awaitable = boto_crud_call(
                        identity_center_client.update_permission_set,
                        InstanceArn=instance_arn,
                        PermissionSetArn=permission_set_arn,
                        **update_resource_params,
                    )
                    tasks.append(
                        plugin_apply_wrapper(apply_awaitable, proposed_changes)
                    )
                else:
                    log.info(log_str, **update_resource_log_params)
                    account_change_details.proposed_changes.extend(proposed_changes)
        else:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_id=name,
                    resource_type=self.resource_type,
                )
            ]
            log_str = "New resource found in code."
            if not ctx.execute:
                # Exit now because apply functions won't work if resource doesn't exist
                log.debug(log_str, **log_params)
                account_change_details.extend_changes(proposed_changes)
                return account_change_details

            log.debug(f"{log_str} Creating resource...", **log_params)

            try:
                permission_set = await boto_crud_call(
                    identity_center_client.create_permission_set,
                    Name=name,
                    InstanceArn=instance_arn,
                    **{
                        param: template_permission_set.get(param)
                        for param in [
                            "Description",
                            "RelayState",
                            "SessionDuration",
                            "Tags",
                        ]
                        if template_permission_set.get(param)
                    },
                )
            except Exception as e:
                for change in proposed_changes:
                    change.exceptions_seen.append(str(e))

            account_change_details.extend_changes(proposed_changes)

            if account_change_details.exceptions_seen:
                log.error(
                    "Unable to create resource on account",
                    exceptions_seen=[
                        cd.exceptions_seen
                        for cd in account_change_details.exceptions_seen
                    ],
                    **log_params,
                )
                return account_change_details

            permission_set_arn = permission_set["PermissionSet"]["PermissionSetArn"]

        tasks.extend(
            [
                apply_permission_set_aws_managed_policies(
                    identity_center_client,
                    instance_arn,
                    permission_set_arn,
                    [
                        mp["Arn"]
                        for mp in template_permission_set.get("ManagedPolicies", [])
                    ],
                    [
                        mp["Arn"]
                        for mp in current_permission_set.get("ManagedPolicies", [])
                    ],
                    log_params,
                ),
                apply_permission_set_customer_managed_policies(
                    identity_center_client,
                    instance_arn,
                    permission_set_arn,
                    template_permission_set.get("CustomerManagedPolicyReferences", []),
                    current_permission_set.get("CustomerManagedPolicyReferences", []),
                    log_params,
                ),
                apply_permission_set_inline_policy(
                    identity_center_client,
                    instance_arn,
                    permission_set_arn,
                    template_permission_set.get("InlinePolicy", "{}"),
                    current_permission_set.get("InlinePolicy", "{}"),
                    log_params,
                ),
                apply_permission_set_permission_boundary(
                    identity_center_client,
                    instance_arn,
                    permission_set_arn,
                    template_permission_set.get("PermissionsBoundary", {}),
                    current_permission_set.get("PermissionsBoundary", {}),
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            account_change_details.extend_changes(
                list(chain.from_iterable(changes_made))
            )

        # apply_account_assignments is a dedicated call due to the request limit on CreateAccountAssignment
        #   per https://docs.aws.amazon.com/singlesignon/latest/userguide/limits.html
        account_assignment_changes = await apply_account_assignments(
            identity_center_client,
            instance_arn,
            permission_set_arn,
            template_account_assignments,
            current_account_assignments,
            log_params,
        )
        if account_assignment_changes:
            account_change_details.extend_changes(account_assignment_changes)

        if ctx.execute and not account_change_details.exceptions_seen:
            if any(changes_made) and not self.deleted:
                try:
                    res = await boto_crud_call(
                        identity_center_client.provision_permission_set,
                        InstanceArn=instance_arn,
                        PermissionSetArn=permission_set_arn,
                        TargetType="ALL_PROVISIONED_ACCOUNTS",
                    )

                    request_id = res["PermissionSetProvisioningStatus"]["RequestId"]

                    for _ in range(20):
                        provision_status = await boto_crud_call(
                            identity_center_client.describe_permission_set_provisioning_status,
                            InstanceArn=instance_arn,
                            ProvisionPermissionSetRequestId=request_id,
                        )

                        if (
                            provision_status["PermissionSetProvisioningStatus"][
                                "Status"
                            ]
                            != "IN_PROGRESS"
                        ):
                            break

                        await asyncio.sleep(1)
                        continue
                except Exception as err:
                    log.warning(
                        "Unable to resolve status when provisioning permission set.",
                        error=str(err),
                        **log_params,
                    )
            log.debug(
                "Successfully finished execution on account for resource",
                changes_made=bool(account_change_details.proposed_changes),
                **log_params,
            )
        elif account_change_details.exceptions_seen:
            log.error(
                "Unable to finish execution on account for resource",
                exceptions_seen=[
                    cd.exceptions_seen for cd in account_change_details.exceptions_seen
                ],
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift on account for resource",
                requires_changes=bool(account_change_details.proposed_changes),
                **log_params,
            )

        return account_change_details

    async def apply(self, config: AWSConfig) -> TemplateChangeDetails:  # noqa: C901
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.resource_id,
            resource_type=self.resource_type,
            template_path=self.file_path,
            exceptions_seen=[],
        )
        log_params = dict(
            resource_type=self.resource_type, resource_id=self.resource_id
        )
        relevant_accounts = []

        for account in config.accounts:
            if not account.identity_center_details:
                continue

            if evaluate_on_provider(self, account):
                relevant_accounts.append(account)
                tasks.append(self._apply_to_account(account))

        if not relevant_accounts:
            if ctx.execute:
                if self.deleted:
                    log_str = "Successfully removed resource."
                    self.delete()
                else:
                    log_str = "No changes detected for resource."
                log.info(log_str, **log_params)

            return template_changes

        if ctx.execute:
            log_str = "Applying changes to resource."
        else:
            log_str = "Detecting changes for resource."

        relevant_accounts_str = [str(account) for account in relevant_accounts]
        log.info(log_str, accounts=relevant_accounts_str, **log_params)

        account_changes = await asyncio.gather(*tasks, return_exceptions=True)
        proposed_changes: list[AccountChangeDetails] = []
        exceptions_seen = set()

        for account_change in account_changes:
            if isinstance(account_change, AccountChangeDetails):
                proposed_changes.append(account_change)
            else:
                exceptions_seen.add(str(account_change))

        if exceptions_seen:
            exceptions_seen = list(exceptions_seen)
            proposed_change_accounts = set(
                change.account for change in proposed_changes
            )
            for aws_account in relevant_accounts:
                if str(aws_account) in proposed_change_accounts:
                    continue
                proposed_changes.append(
                    AccountChangeDetails(
                        account=str(aws_account),
                        resource_id=self.resource_id,
                        resource_type=self.resource_type,
                        exceptions_seen=exceptions_seen,
                    )
                )

        template_changes.extend_changes(proposed_changes)

        if template_changes.exceptions_seen:
            if self.deleted:
                cmd_verb = "removing"
            elif ctx.execute:
                cmd_verb = "applying"
            else:
                cmd_verb = "detecting"
            log_str = f"Error encountered when {cmd_verb} resource changes."
        elif account_changes and ctx.execute:
            if self.deleted:
                self.delete()
                log_str = "Successfully removed resource."
            else:
                log_str = "Successfully applied resource changes."
        elif account_changes:
            log_str = "Successfully detected required resource changes."
        else:
            log_str = "No changes detected for resource."

        log.info(log_str, accounts=relevant_accounts_str, **log_params)
        return template_changes

    @property
    def included_children(self):
        return []

    def set_included_children(self, value):
        pass

    @property
    def excluded_children(self):
        return []

    def set_excluded_children(self, value):
        pass

    @property
    def included_parents(self):
        return self.included_orgs

    def set_included_parents(self, value):
        self.included_orgs = value

    @property
    def excluded_parents(self):
        return self.excluded_orgs

    def set_excluded_parents(self, value):
        self.excluded_orgs = value



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from iambic.core.models import (
    BaseModel,
    BaseTemplate,
    ExpiryModel,
    TemplateChangeDetails,
)

EXAMPLE_LOCAL_DATABASE_TEMPLATE_TYPE = "NOQ::Example::LocalDatabase"

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.example.iambic_plugin import ExampleConfig


class ExampleLocalDatabaseTemplateProperties(BaseModel):
    name: str = Field(..., description="name of Local Database")

    @property
    def resource_type(self) -> str:
        return "example:local_database:properties"

    @property
    def resource_id(self) -> str:
        return self.name


class ExampleLocalDatabaseTemplate(BaseTemplate, ExpiryModel):
    template_type = EXAMPLE_LOCAL_DATABASE_TEMPLATE_TYPE
    properties: ExampleLocalDatabaseTemplateProperties = Field(
        ..., description="Properties for Example Local Database Template"
    )
    name: str = Field(..., description="name of Local Database")

    @property
    def resource_type(self) -> str:
        return "example:local_database"

    @property
    def resource_id(self) -> str:
        return self.name

    async def apply(self, config: ExampleConfig) -> TemplateChangeDetails:
        template_changes = TemplateChangeDetails(
            resource_id=self.resource_id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        template_changes.proposed_changes = []
        return template_changes



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, Any, List, Optional

import okta.models as models

from iambic.core.context import ctx
from iambic.core.exceptions import RateLimitException
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import GlobalRetryController
from iambic.plugins.v0_1_0.okta.exceptions import UserProfileNotUpdatableYet
from iambic.plugins.v0_1_0.okta.models import User
from iambic.plugins.v0_1_0.okta.utils import generate_user_profile, handle_okta_fn

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaOrganization
    from iambic.plugins.v0_1_0.okta.user.models import OktaUserTemplate


async def create_user(
    user_template: OktaUserTemplate,
    okta_organization: OktaOrganization,
) -> Optional[User]:
    """
    Create a new user in Okta.

    Args:
        username (str): The username of the user to create.
        attributes (Dict[str, Any]): The attributes for the user.
        okta_organization (OktaOrganization): The Okta organization to create the user in.

    Returns:
        User: The created User object.
    """

    # TODO: Need ProposedChanges, support context.execute = False
    client = await okta_organization.get_okta_client()

    user_model = {
        "profile": user_template.properties.profile,
    }

    # Create the user
    if ctx.execute:
        async with GlobalRetryController(
            fn_identifier="okta.create_user"
        ) as retry_controller:
            fn = functools.partial(client.create_user, user_model)
            user, _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception(f"Error creating user: {err}")
        return await get_user(
            username=user.profile.login,
            user_id=user.id,
            okta_organization=okta_organization,
        )
    return None


async def get_user(
    username: str,
    user_id: Optional[str],
    okta_organization: OktaOrganization,
) -> Optional[User]:
    """
    Retrieve a user from Okta by username.

    Args:
        username (str): The username of the user to retrieve.
        okta_organization (OktaOrganization): The Okta organization to retrieve the user from.

    Returns:
        User: The retrieved User object.
    """

    client = await okta_organization.get_okta_client()
    if user_id:
        async with GlobalRetryController(
            fn_identifier="okta.get_user"
        ) as retry_controller:
            fn = functools.partial(client.get_user, user_id)
            user, _, err = await retry_controller(handle_okta_fn, fn)
    else:
        async with GlobalRetryController(
            fn_identifier="okta.get_user"
        ) as retry_controller:
            fn = functools.partial(client.get_user, username)
            user, _, err = await retry_controller(handle_okta_fn, fn)
    if err:
        if isinstance(err, asyncio.exceptions.TimeoutError):
            raise err
        if err.error_code == "E0000007":
            return None  # No user exists
        raise Exception(f"Error retrieving user: {err}")

    return User(
        user_id=user.id,
        idp_name=okta_organization.idp_name,
        username=user.profile.login,
        status=user.status.value.lower(),
        extra=dict(
            created=user.created,
        ),
        profile=await generate_user_profile(user),
    )


async def change_user_status(
    user: User,
    new_status: str,
    okta_organization: OktaOrganization,
) -> List[ProposedChange]:
    """
    Change a user's status in Okta.

    Args:
        user (User): The user to change the status of.
        new_status (str): The new status for the user.
        okta_organization (OktaOrganization): The Okta organization to change the user in.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """

    client = await okta_organization.get_okta_client()
    response: list = []

    if user.status == new_status:
        return response

    if user.status:
        current_status = user.status.value
    else:
        current_status = None

    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=user.user_id,
            resource_type=user.resource_type,
            attribute="status",
            current_value=current_status,
            new_value=new_status,
        )
    )

    if ctx.execute:
        async with GlobalRetryController(
            fn_identifier="okta.update_user"
        ) as retry_controller:
            fn = functools.partial(
                client.update_user, user.user_id, {"status": new_status}
            )
            updated_user, _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception("Error updating user status")
        user.status = updated_user.status
    return response


async def update_user_profile(
    proposed_user: OktaUserTemplate,
    user: User,
    new_profile: dict[str, Any],
    okta_organization: OktaOrganization,
    log_params: dict[str, Any],
) -> List[ProposedChange]:
    """
    Update a user's profile in Okta.

    Args:
        user (User): The user to update the profile of.
        new_profile (dict): The new profile for the user.
        okta_organization (OktaOrganization): The Okta organization to update the user in.
    """
    response: list = []
    if not user:
        return response
    current_profile: str = user.profile
    if current_profile == new_profile:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=user.user_id,
            resource_type=user.resource_type,
            attribute="profile",
            change_summary={
                "current_profile": current_profile,
                "new_profile": new_profile,
            },
            current_value=current_profile,
            new_value=new_profile,
        )
    )
    if ctx.execute:
        client = await okta_organization.get_okta_client()
        updated_user_obj = models.User({"profile": new_profile})
        async with GlobalRetryController(
            fn_identifier="okta.update_user",
            retry_exceptions=[
                UserProfileNotUpdatableYet,
                TimeoutError,
                asyncio.exceptions.TimeoutError,
                RateLimitException,
            ],
        ) as retry_controller:
            fn = functools.partial(client.update_user, user.user_id, updated_user_obj)
            _, _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            log.error(
                "Error updating user profile",
                error=err,
                user=user.username,
                current_profile=current_profile,
                new_profile=new_profile,
                **log_params,
            )
            if not proposed_user.deleted:
                if err.error_code == "E0000112":
                    raise UserProfileNotUpdatableYet(
                        "Unable to update profile, user is not fully provisioned"
                    )
                log.error("Error updating user profile", error=err, **log_params)
                raise Exception(f"Error updating user profile: {err}")
        log.info("Updated user profile", user=user.username, **log_params)

    return response


async def update_user_status(
    user: User,
    new_status: str,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the status of a user in Okta.

    Args:
        user (User): The user to update the status of.
        new_status (str): The new status for the user.
        okta_organization (OktaOrganization): The Okta organization to update the user in.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list = []
    if not user:
        return response
    current_status: str = user.status.value
    if current_status == new_status:
        return response
    if user.deleted:
        return response

    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=user.user_id,
            resource_type=user.resource_type,
            attribute="status",
            change_summary={
                "current_status": current_status,
                "proposed_status": new_status,
            },
            current_value=current_status,
            new_value=new_status,
        )
    )

    # TODO: refactor to use an state machine approach
    method = "POST"
    base_endpoint = f"/api/v1/users/{user.user_id}"
    if current_status == "suspended" and new_status == "active":
        api_endpoint = f"{base_endpoint}/lifecycle/unsuspend"
    elif current_status == "active" and new_status == "suspended":
        api_endpoint = f"{base_endpoint}/lifecycle/suspend"
    elif new_status == "deprovisioned" and current_status != "deprovisioned":
        api_endpoint = f"{base_endpoint}/lifecycle/deactivate"
    elif current_status in ["staged", "deprovisioned"] and new_status in [
        "active",
        "provisioned",
    ]:
        api_endpoint = f"{base_endpoint}/lifecycle/activate"
    elif current_status in "provisioned" and new_status == "active":
        api_endpoint = f"{base_endpoint}/lifecycle/reactivate"
    elif current_status == "locked_out" and new_status == "active":
        api_endpoint = f"{base_endpoint}/lifecycle/unlock"
    elif new_status == "recovery":
        api_endpoint = f"{base_endpoint}/lifecycle/reset_password"
    elif new_status == "password_expired":
        api_endpoint = f"{base_endpoint}/lifecycle/expire_password"
    elif new_status == "deleted":
        api_endpoint = f"{base_endpoint}"
        method = "DELETE"
    else:
        log.error(
            "Error updating user status",
            user=user.username,
            current_status=current_status,
            new_status=new_status,
            **log_params,
        )
        raise Exception(
            f"Error updating user status. Invalid transition from {current_status} to {new_status}"
        )

    if ctx.execute:
        client = await okta_organization.get_okta_client()
        request, error = await client.get_request_executor().create_request(
            method=method, url=api_endpoint, body={}, headers={}, oauth=False
        )
        if error:
            log.error(
                "Error updating user status",
                error=error,
                user=user.username,
                current_status=current_status,
                new_status=new_status,
                **log_params,
            )
            raise Exception("Error updating user status")
        okta_response, error = await client.get_request_executor().execute(
            request, None
        )
        if error:
            log.error(
                "Error updating user status",
                error=error,
                user=user.username,
                current_status=current_status,
                new_status=new_status,
                **log_params,
            )
            raise Exception(
                f"Error updating user profile: {error}, "
                f"current_status: {current_status}, "
                f"new_status: {new_status}"
            )
        if okta_response:
            response_body = client.form_response_body(okta_response.get_body())
            log.info(
                "Received Response from Okta: ",
                response_body=response_body,
                **log_params,
            )
    return response


async def maybe_deprovision_user(
    delete: bool,
    user: User,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Delete a user in Okta.

    Args:
        user (User): The user to delete.
        okta_organization (OktaOrganization): The Okta organization to delete the user from.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = []
    if not user:
        return response
    if not delete:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.DELETE,
            resource_id=user.user_id,
            resource_type=user.resource_type,
            attribute="user",
            change_summary={"user": user.username},
            current_value=user.username,
            new_value=None,
        )
    )
    if ctx.execute:
        client = await okta_organization.get_okta_client()
        async with GlobalRetryController(
            fn_identifier="okta.deactivate_or_delete_user"
        ) as retry_controller:
            fn = functools.partial(client.deactivate_or_delete_user, user.user_id)
            _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            if err.error_code == "E0000007":
                log.info(
                    "User already deleted",
                    user=user.username,
                    **log_params,
                )
                return response
            log.error(
                "Error deleting user",
                error=err,
                **log_params,
            )
            raise Exception("Error deleting user")
    return response



from __future__ import annotations

import itertools
import os.path
import uuid

from git import Repo

from iambic.config.dynamic_config import load_config
from iambic.core.git import (
    create_templates_for_deleted_files,
    create_templates_for_modified_files,
    retrieve_git_changes,
)
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import BaseTemplate, ExecutionMessage, TemplateChangeDetails
from iambic.core.parser import load_templates
from iambic.request_handler.expire_resources import flag_expired_resources


async def apply_git_changes(
    config_path: str,
    repo_dir: str,
    allow_dirty: bool = False,
    from_sha: str = None,
    to_sha: str = None,
) -> list[TemplateChangeDetails]:
    """Retrieves files added/updated/or removed when comparing the current branch to master

    Works by taking the diff and adding implied changes to the templates that were modified.
    These are the changes and this function detects:
        Deleting a file
        Removing 1 or more aws_accounts from included_accounts
        Adding 1 or more aws_accounts to excluded_accounts

    :param config_path:
    :param repo_dir:
    :param context:
    :param allow_dirty:
    :param from_sha:
    :param to_sha:
    :return:
    """

    config = await load_config(config_path)
    file_changes = await retrieve_git_changes(
        repo_dir, allow_dirty=allow_dirty, from_sha=from_sha, to_sha=to_sha
    )
    if (
        not file_changes["new_files"]
        and not file_changes["modified_files"]
        and not file_changes["deleted_files"]
    ):
        log.info("No changes found.")
        return []

    exe_message = ExecutionMessage(
        execution_id=str(uuid.uuid4()), command=Command.APPLY
    )
    new_templates = load_templates(
        [git_diff.path for git_diff in file_changes["new_files"]]
    )

    deleted_templates = create_templates_for_deleted_files(
        file_changes["deleted_files"]
    )

    modified_templates_doubles = create_templates_for_modified_files(
        config, file_changes["modified_files"]
    )

    # You can only flag expired resources on new/modified-templates
    await flag_expired_resources(
        [
            template.file_path
            for template in itertools.chain(new_templates, modified_templates_doubles)
            if os.path.exists(template.file_path)
        ]
    )

    template_changes = await config.run_apply(
        exe_message,
        itertools.chain(new_templates, deleted_templates, modified_templates_doubles),
    )

    # note modified_templates_exist_in_repo has different entries from create_templates_for_modified_files because
    # create_templates_for_modified_files actually has two template instance per a single modified file
    modified_templates_exist_in_repo = load_templates(
        [git_diff.path for git_diff in file_changes["modified_files"]]
    )
    commit_deleted_templates(
        repo_dir, modified_templates_exist_in_repo, template_changes
    )

    return template_changes


def commit_deleted_templates(
    repo_dir: str, templates: list[BaseTemplate], details: list[TemplateChangeDetails]
):
    repo = Repo(repo_dir)

    # intended to delete
    deleted_template_path_to_template: dict[str, BaseTemplate] = {
        template.file_path: template for template in templates if template.deleted
    }

    for template_detail in details:
        if template_detail.template_path in deleted_template_path_to_template:
            if template_detail.exceptions_seen:
                log_params = {"path": template_detail.template_path}
                log.error(
                    "add_commits_from_delete_templates cannot be deleted due to exceptions in apply",
                    **log_params,
                )
            else:
                deleted_template_path_to_template[
                    template_detail.template_path
                ].delete()

    diff_list = repo.head.commit.diff()
    if len(diff_list) > 0:
        repo.git.commit("-m", "Delete template after successfully delete resources")



from __future__ import annotations

import asyncio
import os.path
from itertools import chain
from typing import TYPE_CHECKING, Any, List, Optional

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    BaseTemplate,
    ExpiryModel,
    TemplateChangeDetails,
)
from iambic.core.utils import NoqSemaphore
from iambic.plugins.v0_1_0.okta.app.utils import (
    get_app,
    maybe_delete_app,
    update_app_assignments,
    update_app_name,
)
from iambic.plugins.v0_1_0.okta.models import App, Assignment, Status

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig, OktaOrganization

OKTA_GET_APP_SEMAPHORE = NoqSemaphore(get_app, 10)
OKTA_APP_TEMPLATE_TYPE = "NOQ::Okta::App"


class AppProperties(ExpiryModel, BaseModel):
    name: str = Field(..., description="Name of the app")
    status: Optional[Status] = Field(None, description="Status of the app")
    id: Optional[str] = Field(
        None, description="Unique App ID for the app. Usually it's {idp-name}-{name}"
    )
    file_path: str = Field(
        "",
        description="Path to the template file",
        exclude=True,
        hidden_from_schema=True,
    )
    description: Optional[str] = Field("", description="Description of the app")
    extra: Any = Field(None, description=("Extra attributes to store"))
    created: Optional[str] = Field("", description="Date the app was created")
    assignments: List[Assignment] = Field([], description="List of assignments")

    @property
    def resource_type(self) -> str:
        return "okta:app"

    @property
    def resource_id(self) -> str:
        return self.id

    @validator("assignments")
    def sort_groups(cls, v: list[Assignment]):
        sorted_v = sorted(v, key=lambda assignment: assignment.resource_id)
        return sorted_v


class OktaAppTemplate(BaseTemplate, ExpiryModel):
    template_type = OKTA_APP_TEMPLATE_TYPE
    properties: AppProperties = Field(..., description="Properties for the Okta App")
    owner: Optional[str] = Field(None, description="Owner of the app")
    idp_name: str = Field(
        ...,
        description="Name of the identity provider that's associated with the group",
    )

    async def apply(self, config: OktaConfig) -> TemplateChangeDetails:
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.properties.id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_name=self.properties.name,
        )

        if self.iambic_managed == IambicManaged.IMPORT_ONLY:
            log_str = "Resource is marked as import only."
            log.info(log_str, **log_params)
            template_changes.proposed_changes = []
            return template_changes

        for okta_organization in config.organizations:
            if self.idp_name != okta_organization.idp_name:
                continue
            if ctx.execute:
                log_str = "Applying changes to resource."
            else:
                log_str = "Detecting changes for resource."
            log.info(log_str, idp_name=okta_organization.idp_name, **log_params)
            tasks.append(self._apply_to_account(okta_organization))

        account_changes = await asyncio.gather(*tasks)
        template_changes.proposed_changes = [
            account_change
            for account_change in account_changes
            if any(account_change.proposed_changes)
        ]

        proposed_changes = [x for x in account_changes if x.proposed_changes]

        if proposed_changes and ctx.execute:
            log.info(
                "Successfully applied all or some resource changes to all Okta organizations. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        elif proposed_changes:
            log.info(
                "Successfully detected all or some required resource changes on all Okta organizations. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        else:
            log.debug("No changes detected for resource on any account.", **log_params)

        return template_changes

    @property
    def resource_id(self) -> str:
        return self.properties.id

    @property
    def resource_type(self) -> str:
        return "okta:app"

    def set_default_file_path(self, repo_dir: str):
        file_name = f"{self.properties.name}.yaml"
        self.file_path = os.path.expanduser(
            os.path.join(repo_dir, f"resources/okta/app/{self.idp_name}/{file_name}")
        )

    def apply_resource_dict(
        self,
        okta_organization: OktaOrganization,
    ):
        return {
            "name": self.properties.name,
            "owner": self.owner,
            "status": self.properties.status,
            "idp_name": self.idp_name,
            "description": self.properties.description,
            "extra": self.properties.extra,
            "created": self.properties.created,
            "assignments": self.properties.assignments,
        }

    async def _apply_to_account(
        self,
        okta_organization: OktaOrganization,
    ) -> AccountChangeDetails:
        proposed_app = self.apply_resource_dict(okta_organization)
        change_details = AccountChangeDetails(
            account=self.idp_name,
            resource_id=self.properties.id,
            resource_type=self.properties.resource_type,
            new_value=proposed_app,
            proposed_changes=[],
        )

        log_params = dict(
            resource_type=self.properties.resource_type,
            resource_id=self.properties.name,
            organization=str(self.idp_name),
        )

        current_app_task = await OKTA_GET_APP_SEMAPHORE.process(
            [{"okta_organization": okta_organization, "app_id": self.properties.id}]
        )

        current_app: Optional[App] = current_app_task[0]
        if current_app:
            change_details.current_value = current_app

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                change_details.new_value = {}
                return change_details

        app_exists = bool(current_app)
        tasks = []

        if not app_exists and not self.deleted:
            log.error(
                "Iambic does not support creating new apps. "
                "Please create the app in Okta and then import it.",
                **log_params,
            )
            return change_details

        if self.deleted:
            log.info(
                "App is marked for deletion. Please delete manually in Okta",
                **log_params,
            )
            return change_details

        tasks.extend(
            [
                update_app_name(
                    current_app,
                    self.properties.name,
                    okta_organization,
                    log_params,
                ),
                update_app_assignments(
                    current_app,
                    [
                        assignment
                        for assignment in self.properties.assignments
                        if not assignment.deleted
                    ],
                    okta_organization,
                    log_params,
                ),
                maybe_delete_app(
                    self.deleted,
                    current_app,
                    okta_organization,
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            change_details.extend_changes(list(chain.from_iterable(changes_made)))

        if ctx.execute:
            log.debug(
                "Successfully finished execution for resource",
                changes_made=bool(change_details.proposed_changes),
                **log_params,
            )
            # TODO: Check if deleted, remove git commit the change to ratify it
            if self.deleted:
                self.delete()
            self.write()
        else:
            log.debug(
                "Successfully finished scanning for drift for resource",
                requires_changes=bool(change_details.proposed_changes),
                **log_params,
            )

        return change_details


async def get_app_template(okta_app) -> OktaAppTemplate:
    """Get a template for an app"""
    file_name = f"{okta_app.name}.yaml"
    app = OktaAppTemplate(
        file_path=f"resources/okta/apps/{okta_app.idp_name}/{file_name}",
        template_type="NOQ::Okta::App",
        idp_name=okta_app.idp_name,
        properties=AppProperties(
            name=okta_app.name,
            status=okta_app.status,
            id=okta_app.id,
            file_path="{}.yaml".format(okta_app.name),
            assignments=okta_app.assignments,
        ),
    )
    return app



from __future__ import annotations

import itertools
import os
from collections import defaultdict
from typing import TYPE_CHECKING

import aiofiles

from iambic.core import noq_json as json
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
)
from iambic.core.utils import NoqSemaphore, normalize_dict_keys, resource_file_upsert
from iambic.plugins.v0_1_0.aws.event_bridge.models import ManagedPolicyMessageDetails
from iambic.plugins.v0_1_0.aws.iam.policy.models import (
    AWS_MANAGED_POLICY_TEMPLATE_TYPE,
    AwsIamManagedPolicyTemplate,
    ManagedPolicyProperties,
)
from iambic.plugins.v0_1_0.aws.iam.policy.utils import (
    get_managed_policy_across_accounts,
    list_managed_policies,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.template_generation import (
    base_group_str_attribute,
    group_dict_attribute,
    group_int_or_str_attribute,
)
from iambic.plugins.v0_1_0.aws.utils import (
    calculate_import_preference,
    get_aws_account_map,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

RESOURCE_DIR = ["iam", "managed_policy"]


def get_response_dir(exe_message: ExecutionMessage, aws_account: AWSAccount) -> str:
    if exe_message.provider_id:
        return exe_message.get_directory(*RESOURCE_DIR)
    else:
        return exe_message.get_directory(aws_account.account_id, *RESOURCE_DIR)


def get_template_dir(base_dir: str) -> str:
    return str(os.path.join(base_dir, "resources", "aws", *RESOURCE_DIR))


def get_templated_managed_policy_file_path(
    managed_policy_dir: str,
    policy_name: str,
    included_accounts: list[str],
    account_map: dict[str, AWSAccount],
):
    if len(included_accounts) > 1:
        separator = "multi_account"
    elif included_accounts == ["*"] or included_accounts is None:
        separator = "all_accounts"
    else:
        separator = included_accounts[0]
    file_name = (
        policy_name.replace(" ", "")
        .replace("{{var.", "")
        .replace("{{", "")
        .replace("}}_", "_")
        .replace("}}", "_")
        .replace(".", "_")
        .lower()
    )
    return str(os.path.join(managed_policy_dir, separator, f"{file_name}.yaml"))


async def generate_account_managed_policy_resource_files(
    exe_message: ExecutionMessage,
    aws_account: AWSAccount,
) -> dict:
    account_resource_dir = get_response_dir(exe_message, aws_account)
    resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []

    response = dict(account_id=aws_account.account_id, managed_policies=[])
    iam_client = await aws_account.get_boto3_client("iam")
    account_managed_policies = await list_managed_policies(iam_client)

    log.debug(
        "Retrieved AWS IAM Managed Policies.",
        account_id=aws_account.account_id,
        account_name=aws_account.account_name,
        managed_policy_count=len(account_managed_policies),
    )

    for managed_policy in account_managed_policies:
        policy_path = os.path.join(
            account_resource_dir, f'{managed_policy["PolicyName"]}.json'
        )
        response["managed_policies"].append(
            {
                "file_path": policy_path,
                "policy_name": managed_policy["PolicyName"],
                "arn": managed_policy["Arn"],
                "account_id": aws_account.account_id,
            }
        )
        messages.append(
            dict(
                file_path=policy_path, content_as_dict=managed_policy, replace_file=True
            )
        )

    await resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Managed Policies.",
        account_id=aws_account.account_id,
        managed_policy_count=len(account_managed_policies),
    )

    return response


async def generate_managed_policy_resource_file_for_all_accounts(
    exe_message: ExecutionMessage,
    aws_accounts: list[AWSAccount],
    policy_path: str,
    policy_name: str,
) -> list:
    account_resource_dir_map = {
        aws_account.account_id: get_response_dir(exe_message, aws_account)
        for aws_account in aws_accounts
    }
    mp_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []
    response = []

    mp_across_accounts = await get_managed_policy_across_accounts(
        aws_accounts, policy_path, policy_name
    )
    mp_across_accounts = {k: v for k, v in mp_across_accounts.items() if v}

    log.debug(
        "Retrieved AWS IAM Managed Policy for all accounts.",
        policy_name=policy_name,
        policy_path=policy_path,
        total_accounts=len(mp_across_accounts),
    )

    for account_id, managed_policy in mp_across_accounts.items():
        policy_path = os.path.join(
            account_resource_dir_map[account_id],
            f'{managed_policy["PolicyName"]}.json',
        )

        response.append(
            {
                "file_path": policy_path,
                "policy_name": managed_policy["PolicyName"],
                "arn": managed_policy["Arn"],
                "account_id": account_id,
            }
        )
        messages.append(
            dict(
                file_path=policy_path, content_as_dict=managed_policy, replace_file=True
            )
        )

    await mp_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Managed Policy for all accounts.",
        policy_name=policy_name,
        policy_path=policy_path,
        total_accounts=len(mp_across_accounts),
    )

    return response


async def create_templated_managed_policy(  # noqa: C901
    aws_account_map: dict[str, AWSAccount],
    managed_policy_name: str,
    managed_policy_refs: list[dict],
    managed_policy_dir: str,
    existing_template_map: dict,
    config: AWSConfig,
):
    min_accounts_required_for_wildcard_included_accounts = (
        config.min_accounts_required_for_wildcard_included_accounts
    )
    account_id_to_mp_map = {}
    num_of_accounts = len(managed_policy_refs)
    for managed_policy_ref in managed_policy_refs:
        async with aiofiles.open(managed_policy_ref["file_path"], mode="r") as f:
            content_dict = json.loads(await f.read())
            account_id_to_mp_map[
                managed_policy_ref["account_id"]
            ] = normalize_dict_keys(content_dict)

    # calculate preference based on existing template
    prefer_templatized = calculate_import_preference(
        existing_template_map.get(managed_policy_name)
    )

    # Generate the params used for attribute creation
    template_properties = {"policy_name": managed_policy_name}

    # TODO: Fix identifier it should be something along the lines of v but path can vary by account
    #       f"arn:aws:iam::{account_id}:policy{resource['Path']}{managed_policy_name}"
    template_params = {"identifier": managed_policy_name}
    path_resources = list()
    description_resources = list()
    policy_document_resources = list()
    tag_resources = list()
    for account_id, managed_policy_dict in account_id_to_mp_map.items():
        path_resources.append(
            {
                "account_id": account_id,
                "resources": [{"resource_val": managed_policy_dict["path"]}],
            }
        )
        policy_document_resources.append(
            {
                "account_id": account_id,
                "resources": [{"resource_val": managed_policy_dict["policy_document"]}],
            }
        )

        if tags := managed_policy_dict.get("tags"):
            tag_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": tag} for tag in tags],
                }
            )

        if description := managed_policy_dict.get("description"):
            description_resources.append(
                {"account_id": account_id, "resources": [{"resource_val": description}]}
            )

    if (
        len(managed_policy_refs) != len(aws_account_map)
        or len(aws_account_map) <= min_accounts_required_for_wildcard_included_accounts
    ):
        template_params["included_accounts"] = [
            aws_account_map[managed_policy_ref["account_id"]].account_name
            for managed_policy_ref in managed_policy_refs
        ]
    else:
        template_params["included_accounts"] = ["*"]

    path = await group_int_or_str_attribute(
        aws_account_map, num_of_accounts, path_resources, "path"
    )
    if path != "/":
        template_properties["path"] = path

    template_properties["policy_document"] = await group_dict_attribute(
        aws_account_map,
        num_of_accounts,
        policy_document_resources,
        True,
        prefer_templatized=prefer_templatized,
    )

    if description_resources:
        template_properties["description"] = await group_int_or_str_attribute(
            aws_account_map, num_of_accounts, description_resources, "description"
        )

    if tag_resources:
        tags = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            tag_resources,
            True,
            prefer_templatized=prefer_templatized,
        )
        if isinstance(tags, dict):
            tags = [tags]
        template_properties["tags"] = tags

    file_path = get_templated_managed_policy_file_path(
        managed_policy_dir,
        managed_policy_name,
        template_params.get("included_accounts"),
        aws_account_map,
    )
    return create_or_update_template(
        file_path,
        existing_template_map,
        managed_policy_name,
        AwsIamManagedPolicyTemplate,
        template_params,
        ManagedPolicyProperties(**template_properties),
        list(aws_account_map.values()),
    )


async def collect_aws_managed_policies(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    iam_template_map: dict,
    detect_messages: list[ManagedPolicyMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)
    if exe_message.provider_id:
        aws_account_map = {
            exe_message.provider_id: aws_account_map[exe_message.provider_id]
        }

    if detect_messages:
        detect_messages = [
            msg
            for msg in detect_messages
            if isinstance(msg, ManagedPolicyMessageDetails)
        ]
        if not detect_messages:
            return

    existing_template_map = iam_template_map.get(AWS_MANAGED_POLICY_TEMPLATE_TYPE, {})

    log.info(
        "Generating AWS managed policy templates. Beginning to retrieve AWS IAM Managed Policies.",
        accounts=list(aws_account_map.keys()),
    )

    if detect_messages:
        aws_accounts = list(aws_account_map.values())
        generate_mp_resource_file_for_all_accounts_semaphore = NoqSemaphore(
            generate_managed_policy_resource_file_for_all_accounts, 50
        )

        tasks = [
            {
                "exe_message": exe_message,
                "aws_accounts": aws_accounts,
                "policy_path": managed_policy.policy_path,
                "policy_name": managed_policy.policy_name,
            }
            for managed_policy in detect_messages
            if not managed_policy.delete
        ]

        # Remove deleted or mark templates for update
        deleted_managed_policies = [
            managed_policy
            for managed_policy in detect_messages
            if managed_policy.delete
        ]
        if deleted_managed_policies:
            for managed_policy in deleted_managed_policies:
                policy_account = aws_account_map.get(managed_policy.account_id)
                if existing_template := existing_template_map.get(
                    managed_policy.policy_name
                ):
                    if len(existing_template.included_accounts) == 1 and (
                        existing_template.included_accounts[0]
                        == policy_account.account_name
                        or existing_template.included_accounts[0]
                        == policy_account.account_id
                    ):
                        # It's the only account for the template so delete it
                        existing_template.delete()
                    else:
                        # There are other accounts for the template so re-eval the template
                        tasks.append(
                            {
                                "exe_message": exe_message,
                                "aws_accounts": aws_accounts,
                                "policy_path": existing_template.properties.path,
                                "policy_name": existing_template.properties.policy_name,
                            }
                        )

        account_mp_list = (
            await generate_mp_resource_file_for_all_accounts_semaphore.process(tasks)
        )
        account_mp_list = list(itertools.chain.from_iterable(account_mp_list))
        account_policy_map = defaultdict(list)
        for account_policy in account_mp_list:
            account_policy_map[account_policy["account_id"]].append(account_policy)
        account_managed_policies = [
            dict(account_id=account_id, managed_policies=account_managed_policies)
            for account_id, account_managed_policies in account_policy_map.items()
        ]
    elif exe_message.provider_id:
        aws_account = aws_account_map[exe_message.provider_id]
        account_managed_policies = [
            (
                await generate_account_managed_policy_resource_files(
                    exe_message, aws_account
                )
            )
        ]
    else:
        generate_account_managed_policy_resource_files_semaphore = NoqSemaphore(
            generate_account_managed_policy_resource_files, 25
        )
        account_managed_policies = (
            await generate_account_managed_policy_resource_files_semaphore.process(
                [
                    {"exe_message": exe_message, "aws_account": aws_account}
                    for aws_account in aws_account_map.values()
                ]
            )
        )

    # Upsert Managed Policies
    messages = []
    for account in account_managed_policies:
        for managed_policy in account["managed_policies"]:
            messages.append(
                {
                    "policy_name": managed_policy["policy_name"],
                    "arn": managed_policy["arn"],
                    "file_path": managed_policy["file_path"],
                    "aws_account": aws_account_map[managed_policy["account_id"]],
                }
            )

    log.info(
        "Finished retrieving managed policy details",
        accounts=list(aws_account_map.keys()),
    )

    account_managed_policy_output = json.dumps(account_managed_policies)
    with open(
        exe_message.get_file_path(*RESOURCE_DIR, file_name_and_extension="output.json"),
        "w",
    ) as f:
        f.write(account_managed_policy_output)


async def generate_aws_managed_policy_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    iam_template_map: dict,
    detect_messages: list[ManagedPolicyMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)
    existing_template_map = iam_template_map.get(AWS_MANAGED_POLICY_TEMPLATE_TYPE, {})
    resource_dir = get_template_dir(base_output_dir)
    account_managed_policies = await exe_message.get_sub_exe_files(
        *RESOURCE_DIR, file_name_and_extension="output.json", flatten_results=True
    )

    log.info("Grouping managed policies")
    # Move everything to required structure
    for account_mp_elem in range(len(account_managed_policies)):
        for mp_elem in range(
            len(account_managed_policies[account_mp_elem]["managed_policies"])
        ):
            policy_name = account_managed_policies[account_mp_elem]["managed_policies"][
                mp_elem
            ].pop("policy_name")
            account_managed_policies[account_mp_elem]["managed_policies"][mp_elem][
                "resource_val"
            ] = policy_name

        account_managed_policies[account_mp_elem][
            "resources"
        ] = account_managed_policies[account_mp_elem].pop("managed_policies", [])

    grouped_managed_policy_map = await base_group_str_attribute(
        aws_account_map, account_managed_policies
    )

    log.debug("Writing templated managed policies")
    all_resource_ids = set()
    for policy_name, policy_refs in grouped_managed_policy_map.items():
        resource_template = await create_templated_managed_policy(
            aws_account_map,
            policy_name,
            policy_refs,
            resource_dir,
            existing_template_map,
            config,
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    if not detect_messages:
        # NEVER call this if messages are passed in because all_resource_ids will only contain those resources
        delete_orphaned_templates(
            list(existing_template_map.values()), all_resource_ids
        )

    log.info("Finished templated managed policy generation")



from __future__ import annotations

from unittest import IsolatedAsyncioTestCase, mock
from unittest.mock import MagicMock

import boto3
from botocore.exceptions import ClientError
from botocore.stub import Stubber
from moto import mock_s3

from iambic.core.iambic_enum import IambicManaged
from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.utils import (
    boto3_retry,
    create_assume_role_session,
    get_aws_account_map,
)


async def async_noop(*args, **kwargs):
    pass


class TestAWSUtils(IsolatedAsyncioTestCase):
    # Test the boto3_retry function with retries
    async def test_boto3_retry_with_error(self):
        # Create a mock S3 bucket
        with mock_s3():
            s3_client = boto3.client("s3", region_name="us-east-1")
            s3_client.create_bucket(Bucket="test-bucket")

            # Use botocore stubber to mock the s3.head_bucket function
            with Stubber(s3_client) as stubber:
                expected_params = {"Bucket": "test-bucket"}

                # Add the exception response and the successful response to the stubber
                stubber.add_client_error(
                    "head_bucket",
                    expected_params={"Bucket": "test-bucket"},
                    service_error_code="NoSuchBucket",
                    service_message="The specified bucket does not exist",
                    http_status_code=400,
                )
                stubber.add_response(
                    "head_bucket", service_response={}, expected_params=expected_params
                )

                # Patch asyncio.sleep to avoid waiting during tests
                with mock.patch("asyncio.sleep", new=async_noop):

                    @boto3_retry
                    async def mock_boto3_function(bucket_name):
                        return s3_client.head_bucket(Bucket=bucket_name)

                    with self.assertRaises(ClientError):
                        await mock_boto3_function("test-bucket")

    # Test the boto3_retry function without retries
    async def test_boto3_retry_without_retries(self):
        # Create a mock S3 bucket
        with mock_s3():
            s3 = boto3.client("s3", region_name="us-east-1")
            s3.create_bucket(Bucket="test-bucket")

            @boto3_retry
            async def mock_boto3_function(bucket_name):
                return s3.head_bucket(Bucket=bucket_name)

            # Test the original boto3 function
            result = await mock_boto3_function("test-bucket")
            self.assertEqual(result["ResponseMetadata"]["HTTPStatusCode"], 200)

    async def test_boto3_retry_with_throttling(self):
        # Create a mock S3 bucket
        with mock_s3():
            s3_client = boto3.client("s3", region_name="us-east-1")
            s3_client.create_bucket(Bucket="test-bucket")

            # Use botocore stubber to mock the s3.head_bucket function
            with Stubber(s3_client) as stubber:
                # Add the throttling exception and the successful response to the stubber
                for i in range(10):
                    stubber.add_client_error(
                        "head_bucket",
                        expected_params={"Bucket": "test-bucket"},
                        service_error_code="Throttling",
                        service_message="Rate exceeded",
                        http_status_code=429,
                    )

                # Patch asyncio.sleep to avoid waiting during tests
                with mock.patch("asyncio.sleep", new=async_noop):

                    @boto3_retry
                    async def mock_boto3_function(bucket_name):
                        return s3_client.head_bucket(Bucket=bucket_name)

                    with self.assertRaises(
                        ClientError
                    ):  # We expect a Throttline error to be raised
                        await mock_boto3_function("test-bucket")


class TestCreateAssumeRoleSession(IsolatedAsyncioTestCase):
    async def test_create_assume_role_session(self):
        # Set up parameters
        assume_role_arn = "arn:aws:iam::123456789012:role/TestRole"
        region_name = "us-east-1"
        external_id = "test-external-id"
        session_name = "iambic"

        # Mock boto3 session
        boto3_session = MagicMock()

        # Mock the STS client
        sts_client = boto3.client("sts", region_name=region_name)
        boto3_session.client.return_value = sts_client

        # Set up STS response
        expected_credentials = {
            "AccessKeyId": "ACCESS_KEY_ID_123456",
            "SecretAccessKey": "SECRET_KEY",
            "SessionToken": "SESSION_TOKEN",
            "Expiration": "2022-01-01T00:00:00Z",
        }
        sts_response = {
            "AssumedRoleUser": {
                "AssumedRoleId": "AROATESTID:TestRole",
                "Arn": "arn:aws:sts::123456789012:assumed-role/TestRole/{}".format(
                    session_name
                ),
            },
            "Credentials": expected_credentials,
            "PackedPolicySize": 123,
        }

        # Use botocore stubber to mock the sts.assume_role function
        with Stubber(sts_client) as stubber:
            stubber.add_response(
                "assume_role",
                service_response=sts_response,
                expected_params={
                    "RoleArn": assume_role_arn,
                    "RoleSessionName": session_name,
                    "ExternalId": external_id,
                },
            )

            # Call the create_assume_role_session function
            assumed_session = await create_assume_role_session(
                boto3_session, assume_role_arn, region_name, external_id, session_name
            )

            actual_creds = assumed_session.get_credentials().get_frozen_credentials()

            # Check if the returned session has the correct credentials
            self.assertEqual(
                actual_creds.access_key,
                expected_credentials["AccessKeyId"],
            )

            self.assertEqual(
                actual_creds.secret_key,
                expected_credentials["SecretAccessKey"],
            )

            self.assertEqual(
                actual_creds.token,
                expected_credentials["SessionToken"],
            )

    async def test_create_assume_role_session_exception(self):
        # Set up parameters
        assume_role_arn = "arn:aws:iam::123456789012:role/TestRole"
        region_name = "us-east-1"
        external_id = "test-external-id"
        session_name = "iambic"

        # Mock boto3 session
        boto3_session = MagicMock()

        # Mock the STS client
        sts_client = boto3.client("sts", region_name=region_name)
        boto3_session.client.return_value = sts_client

        # Use botocore stubber to mock the sts.assume_role function and raise an exception
        with Stubber(sts_client) as stubber:
            stubber.add_client_error(
                "assume_role",
                service_error_code="AccessDenied",
                service_message="Access Denied",
                http_status_code=400,
                expected_params={
                    "RoleArn": assume_role_arn,
                    "RoleSessionName": session_name,
                    "ExternalId": external_id,
                },
            )

            # Call the create_assume_role_session function and expect an exception
            with self.assertRaises(ClientError):
                await create_assume_role_session(
                    boto3_session,
                    assume_role_arn,
                    region_name,
                    external_id,
                    session_name,
                )


class TestGetAWSAccountMap(IsolatedAsyncioTestCase):
    async def test_get_aws_account_map(self):
        # Create example AWSConfig
        accounts = [
            AWSAccount(
                account_id="123456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_2",
            ),
            AWSAccount(
                account_id="223456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_3",
            ),
            AWSAccount(
                account_id="323456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_4",
            ),
            AWSAccount(
                account_id="423456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_5",
            ),
        ]
        config = AWSConfig(accounts=accounts)

        # Call the get_aws_account_map function
        account_map = await get_aws_account_map(config)

        # Check if the returned account map is correct
        self.assertEqual(len(account_map), len(accounts))
        for account in accounts:
            self.assertIn(account.account_id, account_map)
            self.assertEqual(account_map[account.account_id], account)

    async def test_get_aws_account_map_iambic_managed_disabled(self):
        # Create example AWSConfig
        accounts = [
            AWSAccount(
                account_id="123456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_2",
            ),
            AWSAccount(
                account_id="223456789012",
                iambic_managed=IambicManaged.DISABLED,
                account_name="test_account_3",
            ),
            AWSAccount(
                account_id="323456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_4",
            ),
            AWSAccount(
                account_id="423456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_5",
            ),
        ]
        config = AWSConfig(accounts=accounts)

        # Call the get_aws_account_map function
        account_map = await get_aws_account_map(config)

        # Check if the returned account map is correct
        self.assertEqual(len(account_map), len(accounts) - 1)
        for account in accounts:
            if account.iambic_managed != IambicManaged.DISABLED:
                self.assertIn(account.account_id, account_map)
                self.assertEqual(account_map[account.account_id], account)

    async def test_get_aws_account_map_duplicate_account_id(self):
        # Create example AWSConfig
        accounts = [
            AWSAccount(
                account_id="123456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_2",
            ),
            AWSAccount(
                account_id="123456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_2_duplicate",
            ),
            AWSAccount(
                account_id="223456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_3",
            ),
            AWSAccount(
                account_id="323456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_4",
            ),
            AWSAccount(
                account_id="423456789012",
                iambic_managed=IambicManaged.READ_AND_WRITE,
                account_name="test_account_5",
            ),
        ]
        config = AWSConfig()
        config.accounts = accounts

        # Call the get_aws_account_map function and check for ValueError due to duplicate account_id
        with self.assertRaises(ValueError):
            await get_aws_account_map(config)



from __future__ import annotations

import asyncio
import os
from itertools import chain
from typing import TYPE_CHECKING, Any, List, Optional

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    BaseTemplate,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
    TemplateChangeDetails,
)
from iambic.plugins.v0_1_0.okta.group.utils import (
    create_group,
    get_group,
    maybe_delete_group,
    update_group_description,
    update_group_members,
    update_group_name,
)
from iambic.plugins.v0_1_0.okta.models import Group, UserStatus

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig, OktaOrganization

OKTA_GROUP_TEMPLATE_TYPE = "NOQ::Okta::Group"


class UserSimple(BaseModel, ExpiryModel):
    username: str
    status: Optional[UserStatus] = UserStatus.active

    background_check_status: Optional[bool] = Field(
        False, description="Background check status for the group", exclude=True
    )
    created: Optional[str] = Field(
        None, description="Created date for the group", exclude=True
    )
    domain: Optional[str] = Field(
        None, description="Domain for the group", exclude=True
    )
    extra: Any = Field(None, description=("Extra attributes to store"), exclude=True)
    fullname: Optional[str] = Field(
        None, description="Full name for the group", exclude=True
    )
    groups: Optional[List[str]] = Field(
        None, description="Groups for the group", exclude=True
    )
    idp_name: Optional[str] = Field(
        None, description="IDP name for the group", exclude=True
    )
    profile: Optional[Any] = Field(
        None, description="Profile for the group", exclude=True
    )
    status: Optional[UserStatus] = Field(
        None, description="Status for the group", exclude=True
    )
    updated: Optional[str] = Field(
        None, description="Updated date for the group", exclude=True
    )
    user_id: Optional[str] = Field(
        None, description="User ID for the group", exclude=True
    )

    @property
    def resource_type(self) -> str:
        return "okta:user"

    @property
    def resource_id(self) -> str:
        return self.username


class User(UserSimple):
    idp_name: str
    user_id: Optional[str]
    domain: Optional[str]
    fullname: Optional[str]
    created: Optional[str]
    updated: Optional[str]
    groups: Optional[List[str]]
    background_check_status: Optional[bool]
    extra: Any = Field(None, description=("Extra attributes to store"))


class GroupProperties(ExpiryModel, BaseModel):
    name: str = Field(..., description="Name of the group")
    group_id: str = Field(
        "", description="Unique Group ID for the group. Usually it's {idp-name}-{name}"
    )
    file_path: str = Field(
        "", description="File path of the group", exclude=True, hidden_from_schema=True
    )
    description: Optional[str] = Field("", description="Description of the group")
    extra: Any = Field(None, description=("Extra attributes to store"))
    members: List[UserSimple] = Field([], description="Users in the group")
    identifier: Optional[str] = Field(
        None,
        description="Identifier for the group. Usually it's the group name",
        exclude=True,
    )

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"extra", "metadata_commented_dict"}

    @property
    def resource_type(self) -> str:
        return "okta:group"

    @property
    def resource_id(self) -> str:
        return self.group_id

    @validator("members")
    def sort_groups(cls, v: list[UserSimple]):
        sorted_v = sorted(v, key=lambda member: member.username)
        return sorted_v


class OktaGroupTemplate(BaseTemplate, ExpiryModel):
    template_type = OKTA_GROUP_TEMPLATE_TYPE
    owner: Optional[str] = Field(None, description="Owner of the group")
    properties: GroupProperties = Field(
        ..., description="Properties for the Okta Group"
    )
    idp_name: str = Field(
        ...,
        description="Name of the identity provider that's associated with the group",
    )

    async def apply(self, config: OktaConfig) -> TemplateChangeDetails:
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.properties.group_id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_name=self.properties.name,
        )

        if self.iambic_managed == IambicManaged.IMPORT_ONLY:
            log_str = "Resource is marked as import only."
            log.info(log_str, **log_params)
            template_changes.proposed_changes = []
            return template_changes

        for okta_organization in config.organizations:
            if self.idp_name != okta_organization.idp_name:
                continue
            if ctx.execute:
                log_str = "Applying changes to resource."
            else:
                log_str = "Detecting changes for resource."
            log.info(log_str, idp_name=okta_organization.idp_name, **log_params)
            tasks.append(self._apply_to_account(okta_organization))

        account_changes = await asyncio.gather(*tasks)
        template_changes.proposed_changes = [
            account_change
            for account_change in account_changes
            if any(account_change.proposed_changes)
        ]

        proposed_changes = [x for x in account_changes if x.proposed_changes]

        if proposed_changes and ctx.execute:
            log.info(
                "Successfully applied all or some resource changes to all Okta organizations. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        elif proposed_changes:
            log.info(
                "Successfully detected all or some required resource changes on all Okta organizations. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        else:
            log.debug("No changes detected for resource on any account.", **log_params)

        return template_changes

    @property
    def resource_id(self) -> str:
        return self.properties.group_id

    @property
    def resource_type(self) -> str:
        return "okta:group"

    def set_default_file_path(self, repo_dir: str):
        file_name = f"{self.properties.name}.yaml"
        self.file_path = os.path.expanduser(
            os.path.join(
                repo_dir,
                f"resources/okta/group/{self.idp_name}/{file_name}",
            )
        )

    def apply_resource_dict(self, okta_organization: OktaOrganization):
        return {
            "name": self.properties.name,
            "description": self.properties.description,
            "members": self.properties.members,
        }

    async def _apply_to_account(
        self, okta_organization: OktaOrganization
    ) -> AccountChangeDetails:
        proposed_group = self.apply_resource_dict(okta_organization)
        change_details = AccountChangeDetails(
            account=self.idp_name,
            resource_id=self.properties.group_id,
            resource_type=self.properties.resource_type,
            new_value=proposed_group,  # TODO fix
            proposed_changes=[],
        )

        log_params = dict(
            resource_type=self.properties.resource_type,
            resource_id=self.properties.name,
            organization=str(self.idp_name),
        )

        current_group: Optional[Group] = await get_group(
            self.properties.group_id, self.properties.name, okta_organization
        )
        if current_group:
            change_details.current_value = current_group

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                change_details.new_value = {}
                return change_details

        group_exists = bool(current_group)
        tasks = []

        await self.remove_expired_resources()

        if not group_exists and not self.deleted:
            change_details.extend_changes(
                [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_id=self.properties.group_id,
                        resource_type=self.properties.resource_type,
                    )
                ]
            )
            log_str = "New resource found in code."
            if not ctx.execute:
                log.info(log_str, **log_params)
                # Exit now because apply functions won't work if resource doesn't exist
                return change_details

            log_str = f"{log_str} Creating resource..."
            log.info(log_str, **log_params)

            current_group: Group = await create_group(
                group_name=self.properties.name,
                idp_name=self.idp_name,
                description=self.properties.description,
                okta_organization=okta_organization,
            )
            if current_group:
                change_details.current_value = current_group

        # TODO: Support group expansion
        tasks.extend(
            [
                update_group_name(
                    current_group,
                    self.properties.name,
                    okta_organization,
                    log_params,
                ),
                update_group_description(
                    current_group,
                    self.properties.description,
                    okta_organization,
                    log_params,
                ),
                update_group_members(
                    current_group,
                    [
                        member
                        for member in self.properties.members
                        if not member.deleted
                    ],
                    okta_organization,
                    log_params,
                ),
                maybe_delete_group(
                    self.deleted,
                    current_group,
                    okta_organization,
                    log_params,
                ),
                # TODO
                # upgrade_group_application_assignments
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            change_details.extend_changes(list(chain.from_iterable(changes_made)))

        if ctx.execute:
            log.debug(
                "Successfully finished execution for resource",
                changes_made=bool(change_details.proposed_changes),
                **log_params,
            )
            # TODO: Check if deleted, remove git commit the change to ratify it
            if self.deleted:
                self.delete()
            self.write()
        else:
            log.debug(
                "Successfully finished scanning for drift for resource",
                requires_changes=bool(change_details.proposed_changes),
                **log_params,
            )

        return change_details



from __future__ import annotations

import asyncio
import json
from itertools import chain
from typing import Callable, List, Optional, Union

from jinja2 import BaseLoader, Environment
from pydantic import Field, constr, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import plugin_apply_wrapper, sanitize_string
from iambic.plugins.v0_1_0.aws.iam.models import Path
from iambic.plugins.v0_1_0.aws.iam.policy.utils import (
    apply_managed_policy_tags,
    apply_update_managed_policy,
    delete_managed_policy,
    get_managed_policy,
)
from iambic.plugins.v0_1_0.aws.models import (
    ARN_RE,
    AccessModel,
    AWSAccount,
    AWSTemplate,
    Description,
    Tag,
)
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call

AWS_MANAGED_POLICY_TEMPLATE_TYPE = "NOQ::AWS::IAM::ManagedPolicy"


class Principal(BaseModel):
    aws: Optional[Union[str, list[str]]] = None
    service: Optional[Union[str, list[str]]] = None
    canonical_user: Optional[Union[str, list[str]]] = None
    federated: Optional[Union[str, list[str]]] = None

    def _apply_resource_dict(
        self,
        aws_account: AWSAccount = None,
    ) -> dict:
        resource_dict = super(Principal, self)._apply_resource_dict(aws_account)
        if aws_val := resource_dict.pop("aws", resource_dict.pop("Aws", None)):
            resource_dict["AWS"] = aws_val
        return resource_dict

    @property
    def resource_type(self) -> str:
        return "aws:iam:policy_principal"

    @property
    def resource_id(self) -> str:
        return self.service or self.canonical_user or self.federated or self.aws

    @validator("aws", "service", "canonical_user", "federated")
    def sort_tags(cls, v: list[str]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v)
        return sorted_v


class ConditionalStatement(BaseModel):
    for_all_values: Optional[bool] = None
    for_any_values: Optional[bool] = None


class Condition(BaseModel):
    """TODO: Finish this.

    I put in a pin it because it became obvious this will be a large undertaking.
    The problem stems from the way conditions are structured.

    Take string_equals: Optional[dict] = None for example
    For this one attribute it also requires the following:
        for_all_values__string_equals
        for_any_values__string_equals
        for_all_values__string_equals__if_exists
        for_any_values__string_equals__if_exists
    On top of that we still need to add clarity on what the supported keys are.

    For more info on conditions:
    https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_condition.html

    """

    # Key = service/resource_type
    # Value = resource_value , list[resource_value]

    # String conditionals
    string_equals: Optional[dict] = None
    string_not_equals: Optional[dict] = None
    string_equals_ignore_case: Optional[dict] = None
    string_like: Optional[dict] = None
    string_not_like: Optional[dict] = None

    # Numeric Conditionals
    numeric_equals: Optional[dict] = None
    numeric_not_equals: Optional[dict] = None
    numeric_less_than: Optional[dict] = None
    numeric_less_than_equals: Optional[dict] = None
    numeric_greater_than: Optional[dict] = None
    numeric_greater_than_equals: Optional[dict] = None

    # Date Conditionals
    date_equals: Optional[dict] = None
    date_not_equals: Optional[dict] = None
    date_less_than: Optional[dict] = None
    date_less_than_equals: Optional[dict] = None
    date_greater_than: Optional[dict] = None
    date_greater_than_equals: Optional[dict] = None

    # IP Addr Conditionals
    ip_address: Optional[dict] = None
    not_ip_address: Optional[dict] = None

    # ARN Conditionals
    arn_equals: Optional[dict] = None
    arn_like: Optional[dict] = None
    arn_not_equals: Optional[dict] = None
    arn_not_like: Optional[dict] = None

    # Null Conditional
    null: Optional[dict] = None

    # Boolean conditional
    bool: Optional[dict] = None


class PolicyStatement(AccessModel, ExpiryModel):
    effect: str = Field(..., description="Allow | Deny")
    principal: Optional[Union[Principal, str]] = None
    not_principal: Optional[Union[Principal, str]] = None
    action: Optional[Union[List[str], str]] = Field(
        None,
        description="A single regex or list of regexes. "
        "Values are the actions that can be performed on the resources in the policy statement",
        example="dynamodb:list*",
    )
    not_action: Optional[Union[List[str], str]] = Field(
        None,
        description="An advanced policy element that explicitly matches everything except the specified list of actions."
        "DON'T use this with effect: allow in the same statement OR policy",
    )
    resource: Optional[Union[List[str], str]] = Field(
        None,
        description="A single regex or list of regexes. Values specified are the resources the statement applies to",
    )
    not_resource: Optional[Union[List[str], str]] = Field(
        None,
        description="An advanced policy element that explicitly matches every resource except those specified."
        "DON'T use this with effect: allow and action: '*'",
    )
    condition: Optional[dict] = Field(
        None,
        description="An optional set of conditions to determine of the policy applies to a resource.",
    )
    sid: Optional[str] = Field(
        None,
        description="The Policy Statement ID.",
    )

    @property
    def resource_type(self):
        return "aws:iam:policy_statement"

    @property
    def resource_id(self):
        return self.sid


class AssumeRolePolicyDocument(AccessModel):
    version: str = "2008-10-17"
    statement: Optional[List[PolicyStatement]] = None

    @property
    def resource_type(self) -> str:
        return "aws:iam:assume_role_policy_document"

    @property
    def resource_id(self) -> str:
        return ""


class PolicyDocument(AccessModel, ExpiryModel):
    policy_name: str = Field(
        description="The name of the policy.",
    )
    version: Optional[str]
    statement: Optional[List[PolicyStatement]] = Field(
        None,
        description="List of policy statements",
    )

    @property
    def resource_type(self):
        return "aws:policy_document"

    @property
    def resource_id(self):
        return self.policy_name


class ManagedPolicyDocument(AccessModel):
    version: Optional[str] = None
    statement: Optional[List[PolicyStatement]] = Field(
        None,
        description="List of policy statements",
    )

    def _apply_resource_dict(self, aws_account: AWSAccount = None) -> str:
        resource_dict = super()._apply_resource_dict(aws_account)
        return json.dumps(resource_dict)

    def apply_resource_dict(self, aws_account: AWSAccount) -> dict:
        response = json.loads(self._apply_resource_dict(aws_account))
        variables = {var.key: var.value for var in aws_account.variables}
        variables["account_id"] = aws_account.account_id
        variables["account_name"] = aws_account.account_name

        rtemplate = Environment(loader=BaseLoader()).from_string(json.dumps(response))
        valid_characters_re = r"[\w_+=,.@-]"
        variables = [sanitize_string(v, valid_characters_re) for v in variables]
        data = rtemplate.render(**variables)
        return json.loads(data)

    @property
    def resource_type(self):
        return "aws:iam:managed_policy:policy_document"

    @property
    def resource_id(self):
        return "N/A"


class ManagedPolicyProperties(BaseModel):
    policy_name: str = Field(
        description="The name of the policy.",
    )
    path: Optional[Union[str, List[Path]]] = "/"
    description: Optional[Union[str, list[Description]]] = Field(
        "",
        description="Description of the role",
    )
    policy_document: Union[ManagedPolicyDocument, List[ManagedPolicyDocument]]
    tags: Optional[List[Tag]] = Field(
        [],
        description="List of tags attached to the role",
    )

    @property
    def resource_type(self):
        return "aws:iam:managed_policy:properties"

    @property
    def resource_id(self):
        return self.policy_name

    @validator("path")
    def sort_path(cls, v: list[Path]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("description")
    def sort_description(cls, v: list[Description]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @validator("policy_document")
    def sort_policy_document(cls, v: list[ManagedPolicyDocument]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @classmethod
    def sort_func(cls, attribute_name: str) -> Callable:
        def _sort_func(obj):
            return f"{getattr(obj, attribute_name)}!{obj.access_model_sort_weight()}"

        return _sort_func

    @validator("tags")
    def sort_tags(cls, v: list[Tag]):
        sorted_v = sorted(v, key=cls.sort_func("key"))
        return sorted_v


class AwsIamManagedPolicyTemplate(AWSTemplate, AccessModel):
    template_type = AWS_MANAGED_POLICY_TEMPLATE_TYPE
    properties: ManagedPolicyProperties = Field(
        description="The properties of the managed policy",
    )

    def get_arn_for_account(self, aws_account: AWSAccount) -> str:
        path = self.get_attribute_val_for_account(aws_account, "properties.path", False)
        policy_name = self.properties.policy_name
        return f"arn:{aws_account.partition.value}:iam::{aws_account.account_id}:policy{path}{policy_name}"

    def _apply_resource_dict(self, aws_account: AWSAccount = None) -> dict:
        resource_dict = super()._apply_resource_dict(aws_account)
        resource_dict["Arn"] = self.get_arn_for_account(aws_account)
        return resource_dict

    async def _apply_to_account(self, aws_account: AWSAccount) -> AccountChangeDetails:
        client = await aws_account.get_boto3_client("iam")
        account_policy = self.apply_resource_dict(aws_account)
        policy_name = account_policy["PolicyName"]
        account_change_details = AccountChangeDetails(
            account=str(aws_account),
            resource_id=policy_name,
            resource_type=self.resource_type,
            new_value=dict(**account_policy),
            proposed_changes=[],
            exceptions_seen=[],
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_id=policy_name,
            account=str(aws_account),
        )
        policy_arn = account_policy.pop("Arn")
        current_policy = await get_managed_policy(client, policy_arn)
        if current_policy:
            account_change_details.current_value = {**current_policy}

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                account_change_details.new_value = {}
                return account_change_details

        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        if isinstance(deleted, list):
            deleted = deleted[0].deleted

        if deleted:
            if current_policy:
                account_change_details.new_value = None
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.DELETE,
                        resource_id=policy_name,
                        resource_type=self.resource_type,
                    )
                ]
                log_str = "Active resource found with deleted=false."
                if ctx.execute:
                    log_str = f"{log_str} Deleting resource..."
                log.debug(log_str, **log_params)

                if ctx.execute:
                    apply_awaitable = delete_managed_policy(
                        client, policy_arn, log_params
                    )
                    proposed_changes = await plugin_apply_wrapper(
                        apply_awaitable, proposed_changes
                    )

                account_change_details.extend_changes(proposed_changes)

            return account_change_details

        if current_policy:
            tasks = [
                apply_update_managed_policy(
                    client,
                    policy_arn,
                    json.loads(account_policy["PolicyDocument"]),
                    current_policy["PolicyDocument"],
                    log_params,
                ),
                apply_managed_policy_tags(
                    client,
                    policy_arn,
                    account_policy.get("Tags", []),
                    current_policy.get("Tags", []),
                    log_params,
                ),
            ]

            changes_made: list[list[ProposedChange]] = await asyncio.gather(*tasks)
            if any(changes_made):
                account_change_details.extend_changes(
                    list(chain.from_iterable(changes_made))
                )

        else:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_id=policy_name,
                    resource_type=self.resource_type,
                )
            ]
            log_str = "New resource found in code."
            if not ctx.execute:
                # Exit now because apply functions won't work if resource doesn't exist
                log.debug(log_str, **log_params)
                account_change_details.extend_changes(proposed_changes)
                return account_change_details

            log.debug(f"{log_str} Creating resource...", **log_params)
            if isinstance(account_policy["PolicyDocument"], dict):
                account_policy["PolicyDocument"] = json.dumps(
                    account_policy["PolicyDocument"]
                )
            apply_awaitable = boto_crud_call(client.create_policy, **account_policy)
            account_change_details.extend_changes(
                await plugin_apply_wrapper(apply_awaitable, proposed_changes)
            )

        if ctx.execute and not account_change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution on account for resource",
                changes_made=bool(account_change_details.proposed_changes),
                **log_params,
            )
        elif account_change_details.exceptions_seen:
            log.error(
                "Unable to finish execution on account for resource",
                exceptions_seen=[
                    cd.exceptions_seen for cd in account_change_details.exceptions_seen
                ],
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift on account for resource",
                requires_changes=bool(account_change_details.proposed_changes),
                **log_params,
            )

        return account_change_details


class ManagedPolicyRef(AccessModel, ExpiryModel):
    policy_arn: constr(regex=ARN_RE)
    policy_name: Optional[str] = Field(exclude=True)

    @property
    def resource_type(self):
        return "aws:iam:managed_policy"

    @property
    def resource_id(self):
        return self.policy_arn



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from iambic.core import noq_json as json
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
    get_existing_template_map,
)
from iambic.plugins.v0_1_0.okta.group.models import (
    OKTA_GROUP_TEMPLATE_TYPE,
    GroupProperties,
    OktaGroupTemplate,
)
from iambic.plugins.v0_1_0.okta.group.utils import list_all_groups

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig


def get_resource_dir_args() -> list:
    return ["group"]


def get_response_dir(exe_message: ExecutionMessage) -> str:
    return exe_message.get_directory(*get_resource_dir_args(), "templates")


async def update_or_create_group_template(
    discovered_template: OktaGroupTemplate, existing_template_map: dict
) -> OktaGroupTemplate:
    return create_or_update_template(
        discovered_template.file_path,
        existing_template_map,
        discovered_template.resource_id,
        OktaGroupTemplate,
        {"idp_name": discovered_template.idp_name},
        discovered_template.properties,
        [],
    )


async def collect_org_groups(exe_message: ExecutionMessage, config: OktaConfig):
    assert exe_message.provider_id
    base_path = get_response_dir(exe_message)
    okta_organization = config.get_organization(exe_message.provider_id)
    log.info("Beginning to retrieve Okta groups.", organization=exe_message.provider_id)

    groups = await list_all_groups(okta_organization)
    for group in groups:
        okta_group = OktaGroupTemplate(
            file_path="unset",
            idp_name=group.idp_name,
            properties=GroupProperties(
                group_id=group.group_id,
                name=group.name,
                description=group.description,
                members=[json.loads(m.json()) for m in group.members],
            ),
        )
        okta_group.file_path = os.path.join(base_path, f"{group.name}.yaml")
        okta_group.write()

    log.info(
        "Finished retrieving Okta groups.",
        okta_org=exe_message.provider_id,
        group_count=len(groups),
    )


async def generate_group_templates(
    exe_message: ExecutionMessage, output_dir: str, detect_messages: list = None
):
    """List all groups in the domain, along with members and settings"""
    base_path = os.path.expanduser(output_dir)
    existing_template_map = await get_existing_template_map(
        base_path, OKTA_GROUP_TEMPLATE_TYPE
    )
    all_resource_ids = set()

    log.info("Updating and creating Okta group templates.")

    groups = await exe_message.get_sub_exe_files(
        *get_resource_dir_args(), "templates", file_name_and_extension="**.yaml"
    )
    # Update or create templates
    for group in groups:
        group = OktaGroupTemplate(file_path="unset", **group)
        group.set_default_file_path(output_dir)
        resource_template = await update_or_create_group_template(
            group, existing_template_map
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    # Delete templates that no longer exist
    delete_orphaned_templates(list(existing_template_map.values()), all_resource_ids)

    log.info("Finish updating and creating Okta group templates.")
